"""
collection of functions for interacting with UBCS/LCP channel archiver
"""

def get_all_archivers():
    """
    returns all available archivers
    """
    pass

def get_all_pvnames_in_archiver(archiver):
    """
    returns names of all PV names in a give archiver
    """
    pass

def read_pv(pvname, archiver = 'LaserLab', start=None, end=None):
    """
    reads file generated by Channel Archiver and return all values between time minimim (tmin) and time maximum (tmax)
    """
    from sys import path
    path.append('/net/femto/C/All Projects/APS/Instrumentation/Software/Lauecollect/')
    from channel_archiver import channel_archiver

    time, value = channel_archiver(archiver).history(pvname, start, end)
    return time, value

def read_pv_time():
    """np.where(np.array(t)<=tt)[0][-
        np.where(np.array(t)>=tt)[0][0]
        """
    pass

def get_archived_PV(archiver, pvname, timestamps):
    """takes a one process variable name and a list of time stamps and returns corresponding values as a dictionary.

    Parameters
    ----------
    pvname (string)
    timestamps (list)
    exclude (list)

    Returns
    -------
    data (dict)

    Examples
    --------
    """
    from numpy import where, array, diff, mean
    timestamps = array(timestamps)
    dt = mean(diff(timestamps))
    t,value = read_pv(pvname = pvname, archiver = archiver, start=timestamps[0] - +5*dt, end=timestamps[-1]+5*dt)
    from scipy.interpolate import UnivariateSpline
    spl = UnivariateSpline(t, value)
    return timestamps, spl(timestamps)


def get_archived_PVs(archiver, pv_names, timestamps):
    """take a list of process variables and a list of time stamps and returns corresponding values as a dictionary.

    Parameters
    ----------
    pv_names (list)
    timestamps (list)
    exclude (list)

    Returns
    -------
    data (dict)

    Examples
    --------
    """
    result = {}
    for pv_name in pv_names:
        t,value = get_archived_PV(archiver = archiver, pvname = pv_name, timestamps = timestamps)
        result[pv_name] = value
    result['time'] = timestamps
    return result



def new_thread(function, *args, **kwargs):
    """
    launches the input 'function' in a separate thread with daemon == True.

    Explanation: Some threads do background tasks, like sending keepalive packets, or performing periodic garbage collection, or whatever. These are only useful when the main program is running, and it's okay to kill them off once the other, non-daemon, threads have exited.

    Without daemon threads, you'd have to keep track of them, and tell them to exit, before your program can completely quit. By setting them as daemon threads, you can let them run and forget about them, and when your program quits, any daemon threads are killed automatically.
    @Chris Jester-Young https://stackoverflow.com/a/190017/8436767

    Parameters
    ----------
    function:
        function object
    *args:
        iterable of arguments
    **kwargs:
        dictionary of keywords
            daemon:
                flag, if the thread is daemon(True) or not(False).


    Returns
    -------
    thread:
        thread object

    Examples
    --------
    the example of usage

    >>> def func(*args,**kwargs):
            print('keywords',kwargs)
            print('args',args)
            from time import time,sleep
            t1 = time()
            sleep(args[0])
            t2 = time()
            print('this thread slept for {} seconds'.format(t2-t1))
            print('input parameters: args = {}, kwargs = {!r}'.format(args[0],kwargs['keyword']))
    >>> new_thread(func, 1, keyword = 'this is key 1 value')
    this thread slept for 1.002244234085083 seconds
    input parameters: args = 1, kwargs = 'keywords argument'

    """
    from threading import Thread
    thread = Thread(target=function, args = args, kwargs = kwargs)
    try:
        daemon = kwargs['daemon']
    except:
        daemon = True
    thread.daemon = daemon
    thread.start()
    return thread

def start_new_safe_thread(function, daemon = True, *args ,  **kwargs):
    """
    alias for new_thread for back compatibility
    """
    import warnings
    warnings.warn(
            "start_new_safe_thread will be deprecated in version 0.0.7, use new_thread instead",
             PendingDeprecationWarning
        )
    return new_thread(function, args , kwargs)



from .multithreading import *
import warnings

warnings.simplefilter('always', DeprecationWarning)

warnings.warn("The 'ubcs_auxiliary.threading' class was renamed to 'ubcs_auxiliary.multithreading' to make room for 'multithreading' and 'multiprocessing'", DeprecationWarning )



def new_child_process(function,  *args, **kwargs):
    """
    launches the input 'function' in a separate thread with daemon == True.

    Explanation: Some threads do background tasks, like sending keepalive packets, or performing periodic garbage collection, or whatever. These are only useful when the main program is running, and it's okay to kill them off once the other, non-daemon, threads have exited.

    Without daemon threads, you'd have to keep track of them, and tell them to exit, before your program can completely quit. By setting them as daemon threads, you can let them run and forget about them, and when your program quits, any daemon threads are killed automatically.
    @Chris Jester-Young https://stackoverflow.com/a/190017/8436767

    Parameters
    ----------
    function:
        function object
    *args:
        iterable of arguments
    **kwargs:
        dictionary of keywords

    Returns
    -------
    process:
        process object

    Examples
    --------
    the example of usage

    >>> def func(*args,**kwargs):
            print('keywords',kwargs)
            print('args',args)
            from time import time,sleep
            t1 = time()
            sleep(args[0])
            t2 = time()
            print('this thread slept for {} seconds'.format(t2-t1))
            print('input parameters: args = {}, kwargs = {!r}'.format(args[0],kwargs['keyword']))
    >>> new_child_process(func, 1, keyword = 'this is key 1 value')
    this thread slept for 1.002244234085083 seconds
    input parameters: args = 1, kwargs = 'keywords argument'
    """
    from multiprocessing import Process
    p = Process(target=function, args=args, kwargs = kwargs)
    p.daemon = True
    p.start()
    return p

class ParallelProcessing():
    def __init__(self, function, iter, *args, **kwargs):
        """
        This is a specialized wrapper around multiprocessing.Process object. It
        launches multiple separate processes with the same input 'function' but different argument that comes from an iterable object like tuple or list.

        Parameters
        ----------
        function :: function
            function object
        args :: list or tuple
            iterable of arguments
        kwargs :: dictionary
            keyword arguments that will be passed to the function

        Returns
        -------
        jobs :: list
            list of jobs

        Examples
        --------
        >>> mp = ParallelProcessing(function, (1,2,3))
        """
        from multiprocessing import Process
        print(args,kwargs)
        self.jobs = []
        for arg in args:
            p = Process(target=function,args=(arg,), **kwargs)
            p.daemon = True
            p.start()
            self.jobs.append(p)

class MultiProcessing():
    def __init__(self, function, args, **kwargs):
        """
        This is a specialized wrapper around multiprocessing.Process object. It
        launches multiple separate processes with the same input 'function' but different argument that comes from an iterable object like tuple or list.

        Parameters
        ----------
        function :: function
            function object
        args :: list or tuple
            iterable of arguments
        kwargs :: dictionary
            keyword arguments that will be passed to the function

        Returns
        -------
        jobs :: list
            list of jobs

        Examples
        --------
        >>> mp = ParallelProcessing(function, (1,2,3))
        """
        from multiprocessing import Process
        print(args,kwargs)
        self.jobs = []
        for arg in args:
            p = Process(target=function,args=(arg,), **kwargs)
            p.start()
            self.jobs.append(p)
        import warnings
        warnings.warn("MultuProcessing class will be deprecated in future versions. The new name is ParallelProcessing", DeprecationWarning)

def function(N, **kwargs):
    """
    an example function for test purposes. It sleeps for N seconds. Prints statements when sleep starts and finishes.
    """
    from time import time, sleep

    print_text_kwarg(**kwargs)
    print_dict_kwarg(**kwargs)
    print_var_kwarg(**kwargs)

    print('Starting Process')
    t1 = time()
    print('[{} s] Start sleeping for {} seconds'.format(round(time()-t1,4),N))
    print('[{} s] Sleeping... '.format(round(time()-t1,4)))
    sleep(N)
    print('[{} s] Done sleeping for {} seconds'.format(round(time()-t1,4),N))

def print_text_kwarg(**kwargs):
    print(f"kwarg text = {kwargs['text']}")
def print_dict_kwarg(**kwargs):
    print(f"kwarg dict = {kwargs['dict']}")
def print_var_kwarg(**kwargs):
    print(f"kwarg var = {kwargs['var']}")

def test():
    """
    simple test showing how to use
    """
    kwargs = {'text': 'my new text', 'var' : 15.64, 'dict' : {'key':'my new value'}}
    mp = MultiProcessing(function, args = (1,2,3), kwargs = kwargs)

def test2():
    """
    simple test showing how to use
    """
    kwargs = {'text': 'my new text', 'var' : 15.64, 'dict' : {'key':'my new value'}}
    mp = ParallelProcessing(function, args = (1,2,3), kwargs = kwargs)

def test3(root, t = 120):
    from time import time, ctime, sleep
    import os
    filename = os.path.join(root,str(time())+'.txt')
    print(root,t)
    t_start = time()
    while time()-t_start <=t:
        with open(filename,'a') as f:
            f.write(str(ctime(time()))+'\n')
            sleep(1)
    print('the process is done')




from ._version import get_versions
__version__ = get_versions()['version']
del get_versions

from .advsleep import precision_sleep, interupt_sleep
from .save_load_object import save_to_file
from .save_load_object import load_from_file

from . import os
from . import traveling_salesman
from . import numerical
from . import multithreading
from . import multiprocessing
from . import video
from . import channel_archiver
from .other import beep



def get_video_info(filename):
    """
    returns information about video file in a dictionary format


    Parameters
    ----------
    filename :: (string)
        full filename including root

    Returns
    -------
    video_info :: (dictionary)

    Examples
    --------
    >>> video_info = get_video_info(filename)
    """
    import cv2
    dic = {}
    vidcap = cv2.VideoCapture(filename)
    dic['filename'] = filename
    dic['width'] = int(vidcap.get(cv2.CAP_PROP_FRAME_WIDTH))
    dic['height'] = int(vidcap.get(cv2.CAP_PROP_FRAME_HEIGHT))
    dic['gain'] = vidcap.get(cv2.CAP_PROP_GAIN)
    dic['iso'] = vidcap.get(cv2.CAP_PROP_ISO_SPEED)
    dic['frame_count'] = int(vidcap.get(cv2.CAP_PROP_FRAME_COUNT))
    dic['format'] = vidcap.get(cv2.CAP_PROP_FORMAT)
    dic['fps'] = round(vidcap.get(cv2.CAP_PROP_FPS),0)
    return dic

def parse_video_into_frames(filename, verbose = False):
    import cv2
    from os.path import exists,split
    from os import mkdir
    import sys
    root = split(filename)[0]
    prefix = split(filename)[1].split('.')[0]
    print('Checking if the file exist ...')
    print('File does exist:',exists(filename))
    images_root = root +'/'+prefix+'_images/'
    if not exists(images_root):
        mkdir(images_root)
    vidcap = cv2.VideoCapture(filename)
    success,image = vidcap.read()
    count = 0
    success = True
    while success:
      success,image = vidcap.read()
      if success:
          if verbose:
              print(f'extractinf frame # {count} and saving to {images_root}')
          #cv2.imwrite(images_root+"frame%d.tiff" % count, image)     # save frame as JPEG file
          from PIL import Image
          im = Image.fromarray(image)
          im.save(images_root+f"frame{count}.tiff" , dpi = (600,600))
          if cv2.waitKey(10) == 27:                     # exit if Escape is hit
              break
          count += 1
      else:
          count += 1




from ..saved_property import DataBase, SavedProperty

def test_tmpdir(tmpdir):
    print('temp directory is ',tmpdir)

def test_create_a_b(tmpdir):
    """
    Creates a class with default values.
    changes them and destroy itself.
    Create a different instance with the same class and confirm that values were uploaded correctly.
    """
    from random import random
    class A():
        db = DataBase(root = tmpdir, name = 'simpledb')
        a = SavedProperty(db,'a',0.0).init()
        b = SavedProperty(db,'b',0.0).init()
        c = SavedProperty(db,'c',0.0).init()
        d = SavedProperty(db,'d',0.0).init()
    for i in range(100):
        a = A()
        value1 = random()
        value2 = random()
        a.a = value1
        a.b = value1
        assert a.a == a.b
        a.b = value2
        del a
        b = A()
        assert b.a == value1
        assert b.b == value2

def test_arrays(tmpdir):
    from random import random
    from numpy import zeros
    class Array():
        db = DataBase(root = tmpdir, name = 'simpledb')
        array = SavedProperty(db,'array',zeros((100,100))).init()

    a = Array()
    assert a.array.mean() == 0
    a.array += 1
    assert a.array.mean() == 1
    del a
    b = Array()
    assert b.array.mean() == 1



#!/bin/env python
"""
The server LL code for the hierarchical level instrumentation

author: Valentyn Stadnytskyi NIH\LCP
data: Sept 2018 - Nov 20 2018

0.0.2 - code under development
0.0.3 - changed init_server function. It was altering the way how logging was working in higher up modules
"""
__version__ = "0.0.3"
__date__ = "November 20, 2018"



import traceback
from pdb import pm

from numpy import nan, mean, std, nanstd, \
     asfarray, asarray, hstack, array, concatenate, \
     delete, round, vstack, hstack, zeros, transpose, \
     split, unique, nonzero, take, savetxt, min, max
from time import time, sleep, clock
import sys
import os.path

from time import gmtime, strftime, time
from logging import debug,info,warning,error


from ubcs_auxiliary.threading import new_thread

from struct import pack, unpack
from timeit import Timer, timeit

from datetime import datetime

import msgpack
import msgpack_numpy as m
import socket

import platform
server_name = platform.node()

class Server_LL(object):

    def __init__(self, name = '', ports = range(2030,2051)):
        """
        to initialize an instance and create main variables
        """
        if len(name) == 0:
            self.name = 'test_communication_LL'
        else:
            self.name = name
        self.running = False
        self.network_speed = 12**6 # bytes per second
        self.push_subscribe_lst = []
        self.last_call_lst = []

    def init_server(self, name = '', ports = range(2030,2051)):
        '''
        Proper sequence of socket server initialization
        '''
        if len(name) == 0:
            self.name = 'test_communication_LL'
        else:
            self.name = name
        self.sock = self.init_socket(ports = ports)
        self._set_commands()
        if self.sock is not None:
            flag = True
        else:
            flag = False
        if flag:
            self._start()
        info('server init.server')

    def stop(self):
        self.running = False
        self.sock.close()

    def init_socket(self,ports = range(2030,2051)):
        '''
        initializes socket for listening, creates sock and bind to '' with a port somewhere between 2030 and 2050
        '''
        import socket
        for port in ports:
            flag = False
            try:
                sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                sock.bind(('', port))
                flag = True
            except:
                pass
            if flag:
                self.port = port
                try:
                    self.ip_address = socket.gethostbyname(socket.gethostname())
                except:
                    self.ip_address = '127.0.0.1'
                sock.listen(100)
                break
            else:
                sock = None
                self.port = None
        info('server init.socket: port %r' %(port))
        return sock

    def _set_commands(self):
        """
        Set type definition, the dictionary of command excepted by the server
        Standard supported commands:
        - "help"
        - "init"
        - "close"
        - "abort"
        - "snapshot"
        - "subscribe"
        - "task"
        - "controlls"
        - "indicators"
        - "push_updates"
        """
        self.commands = {}
        self.commands[b'help'] = None
        self.commands[b'init'] = None
        self.commands[b'close'] = None
        self.commands[b'abort'] = None
        self.commands[b'snapshot'] = None
        self.commands[b'subscribe'] = None
        self.commands[b'task'] = None
        self.commands[b'controls'] = None
        self.commands[b'indicators'] = None
        self.commands[b'push_updates'] = None


    def set_tasks(self):
        self.tasks = {}
        self.tasks[b'push updates'] = 'push updates'


    def _get_commands(self):
        """
        returns the dictionary with all supported commands
        """
        return self.commands

    def _start(self):
        '''
        creates a separete thread for server_thread function
        '''
        new_thread(self._run)

    def _run(self):
        """
        runs the function _run_once in a while True loop
        """
        self.running = True
        while self.running:
            self._run_once()
        self.running = False

    def _run_once(self):
        """
        creates a listening socket.
        """

        client, addr = self.sock.accept()
        info('Client has connected: %r,%r' %(client,addr))

        try:
            msg_in = self._receive(client)
            debug('msg in %r' % msg_in)
        except:
            msg_in = {b'command':b'unknown',b'message':b'unknown'}
            error('%r, with input message%r' %(traceback.format_exc(),msg_in))
        msg_out = self._receive_handler(msg_in,client)
        self._send(client,msg_out)

        try:
            message = list(msg_in[b'message'].keys())
        except:
            message = msg_in[b'message']
        self._log_last_call(client, addr, command = msg_in[b'command'], message = message)
        info('Client has been served: %r,%r' %(client,addr))


    def _transmit_handler(self,command = '', message = ''):
        from time import time
        res_dic  = {}
        res_dic[b'command'] = command
        res_dic[b'time'] = time()
        res_dic[b'message'] = message
        return res_dic

    def _receive_handler(self,msg_in,client):
        """
        the incoming msg_in has N mandatory fields:  command, message and time
        """
        from time import time
        res_dic = {}
        #the input msg has to be a dictionary. If not, ignore. FIXIT. I don't know how to do it in Python3
        debug('_receive_handler: command received: %r' % msg_in)
        err = ''
        try:
            keys = msg_in.keys()
            command = msg_in[b'command']
            message_in =  msg_in[b'message']
            debug('_receive_handler: message_in %r, command %r' %(message_in,command))
            res_dic[b'command'] = command
            flag = True
            if command in list(self.commands.keys()):
                if self.commands[command] != None:
                    res_dic = self.commands[command](msg_in = message_in,client = client)
                else:
                    res_dic[b'message'] = None
                    err += 'the %r function is not linked' %(command)
            else:
                flag = False
                err += 'the command %r is not supported by the server.' % command
        except:
            error(traceback.format_exc())
            err += traceback.format_exc()
            res_dic[b'command'] = 'unknown'
            res_dic[b'message'] = 'The quote of the day: FIXIT I hope you enjoyed it.'
            res_dic[b'flag'] = False
            res_dic[b'error'] = err
        res_dic[b'time'] = time()
        res_dic[b'error'] = err
        debug('_receive_handler: res_dic %r' % res_dic)
        return res_dic


    def _receive(self,client):
        """
        descritpion:
        client sends 20 bytes with a number of expected package size.
        20 bytes will encode a number up to 10**20 bytes.
        This will be enough for any possible size of the transfer

        input:
        client - socket client object

        output:
        unpacked data
        """
        import msgpack
        import msgpack_numpy as msg_m
        a = client.recv(20)
        length = int(a)
        if length != 0:
            msg_in = ''.encode()
            while len(msg_in) < length:
                msg_in += client.recv(length - len(msg_in))
                if len(msg_in) < length:
                    sleep(0.01)
        else:
            msg_in = ''
        response = msgpack.unpackb(msg_in, object_hook=msg_m.decode)
        return response

    def _send(self,client,msg_out):
        """
        descrition:
        uses msgpack to serialize data and sends it to the client
        """
        debug('command send %r' % msg_out)
        msg = msgpack.packb(msg_out, default=m.encode)
        length = str(len(msg))
        if len(length)!=20:
            length = '0'*(20-len(length)) + length
        try:
            client.sendall(length.encode())
            client.sendall(msg)
            flag = True
        except:
            error(traceback.format_exc())
            flag = False
        return flag

    def _connect(self,ip_address,port):
        server = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        try:
            server.settimeout(1)
            server.connect((ip_address , port))
            server.settimeout(None)
            debug("Connection success!")
        except:
            error('%r' %(traceback.format_exc()))
            server = None
        return server

    def _transmit(self,command = '', message = '' ,ip_address = '127.0.0.1',port = 2030):
        if isinstance(message,dict):
            pass
            #message[b'ip_address'] = ip_address
            #message[b'port'] = port
        else:
            message = {}
            #message[b'ip_address'] = ip_address
            #message[b'port'] = port
        msg_out = self._transmit_handler(command = command, message = message)
        debug("_transmit msg out (%r)" % msg_out)
        client = self._connect(ip_address = ip_address,port = port)
        debug("connected to the client (%r)" % client)
        if client is not None:
            flag = self._send(client,msg_out)
            self.response_arg = self._receive(client)
            self._server_close(client)
        else:
            self.response_arg = None
        return self.response_arg

    def _server_close(self,server):
        server.close()
        return server

    def _log_last_call(self,client = None, addr = None, command = None, message = None):
        from time import time
        self.last_call_lst = [{b'time':time(),b'ip_address':addr[0], b'command':command,b'message':message}]
        #if len(self.last_call_lst) > 10:
            #self.last_call_lst.pop(0)

#***************************************************
#*** wrappers for basic response functions *********
#***************************************************

    def help(self,msg_in  =None, client = None):
        msg = {}
        msg[b'server name']= self.name
        msg[b'server port']= self.port
        msg[b'server ip address']= socket.gethostbyname(socket.gethostname())
        msg[b'server version']= __version__
        msg[b'commands'] = list(self.commands.keys())
        #msg[b'tasks'] = self.dev_get_tasks
        try:
            try:
                getattr(device.controls,'running')
            except:
                err = 'device offline'
            msg[b'device help'] = self.commands[b'help']()
        except:
            error(traceback.format_exc())
            msg[b'controls'] = 'device offline'
            msg[b'indicators'] = 'device offline'
        return msg

    def subscribe(self,client = None,msg_in = {}):
        from time import time
        new_client_dic = {}
        msg = {}
        err = ''
        new_client_dic[b'port'] = msg_in[b'port']
        new_client_dic[b'ip_address'] = client.getpeername()[0]
        new_client_dic[b'controls'] = msg_in[b'controls']
        new_client_dic[b'indicators'] = msg_in[b'indicators']
        try:
            self.push_subscribe_lst.pop(self.push_subscribe_lst.index(new_client_dic))
        except:
            info(traceback.format_exc())
        self.push_subscribe_lst.append(new_client_dic)
        msg[b'flag'] = True
        msg[b'error'] = err
        return msg

    def push_subscribed_updates(self, controls = [], indicators = [], subscriber = -1):
        """
        pushes updates to subscribed clients stored in self.push_subscriber_lst
        """
        if len(controls) == 0:
            debug('server_LL: the list of controls names ie empty. Nothing to push. Indicators: %r' % indicators)
        if len(indicators) == 0:
            debug('server_LL: the list of indicators names ie empty. Nothing to push. Controls: %r' % controls)
        try:
            a = self.push_subscribe_lst[subscriber]
        except:
            debug("subscriber %r in server.push_subscribed_updates doesn't exist" %(subscriber))
            subscriber = -1

        if subscriber == -1:
            push_subscriber_lst = self.push_subscribe_lst
        else:
            push_subscriber_lst = [self.push_subscribe_lst[subscriber]]

        #Stepping through all subscribers in push_subscriber_lst
        for client_dic in push_subscriber_lst:
            indicators_dic = {}
            controls_dic = {}
            msg_dic = {}
            debug("if indicators in %r" % client_dic.keys())

            #checks if client subscribed to indicators
            if b'indicators' in list(client_dic.keys()) and len(indicators) !=0:
                if indicators == [b'all']:
                    for item in client_dic[b'indicators']:
                        indicators_dic[item] = None
                else:
                    for item in indicators:
                        if item in client_dic[b'indicators']:
                            indicators_dic[item] = None
                debug("push_subscribed_updates indicators_dic = %r" % indicators_dic)
                msg_dic[b'indicators']= self.commands[b'indicators'](msg_in = indicators_dic)
            else:
                msg_dic[b'indicators'] = ''


            if b'controls' in list(client_dic.keys()) and len(controls) !=0:
                debug("push_subscribed_updatescontrols = %r" % controls)
                if controls == [b'all']:
                    for item in client_dic[b'controls']:
                        controls_dic[item] = None
                else:
                    for item in controls:
                        if item in client_dic[b'controls']:
                            controls_dic[item] = None
                debug("push_subscribed_updates controls_dic = %r" % controls_dic)
                msg_dic[b'controls'] = self.commands[b'controls'](controls_dic)
            else:
                msg_dic[b'controls'] = ''

            debug('client_dic %r' % client_dic)
            debug('msg = '+str(msg_dic))
            response = 0
            try:
                if len(indicators_dic) != 0:
                    response = self._transmit(command = 'indicators',
                                   message = msg_dic[b'indicators'] ,
                                   ip_address = client_dic[b'ip_address'],
                                   port = client_dic[b'port'])

                if len(controls_dic) != 0:
                    response = self._transmit(command = 'controls',
                                   message = msg_dic[b'controls'] ,
                                   ip_address = client_dic[b'ip_address'],
                                   port = client_dic[b'port'])
            except:
                response = None
                warning(traceback.format_exc())
            if response == None:
                index = self.push_subscribe_lst.index(client_dic)
                self.push_subscribe_lst.pop(index)
                info('The subscribed client\server is not avaiable.')




server = Server_LL()


if __name__ == "__main__": #for testing
    from tempfile import gettempdir
    import logging
    logging.basicConfig(#filename=gettempdir()+'/communication_LL.log',
                        level=logging.DEBUG, format="%(asctime)s %(levelname)s: %(message)s")
    server = Server_LL(name = 'client')
    self = server
    server.init_server()

    info('Server: is running at %r,%r and port %r' %(server.sock.getsockname(),
                                                      socket.gethostbyname(socket.gethostname()),
                                                      server.port))



"""
benchmark file to test conversion of raw byte stream to image format
"""


def benchmark_binarr_to_number():
    from lcp_video.analysis import get_mono12packed_conversion_mask
    from ubcs_auxiliary.save_load_object import load_from_file
    from time import time
    import timeit
    from numpy import array
    preparation= ''
    preparation += "from ubcs_auxiliary.numerical import binarr_to_number;"
    preparation += "from numpy import random;"
    preparation += "vec = random.randint(0,1,(64,))"
    testcode = 'n = binarr_to_number(vec)'
    t = timeit.Timer(testcode,preparation)
    temp = array(t.repeat(40,200))/200
    print(f"{round(temp.mean(),5)} +- {round(temp.std(),6)}")

def benchmark_mono12packed_to_image():
    #mono12packed_to_image(rawdata, height, width, mask)
    from lcp_video.analysis import get_mono12packed_conversion_mask
    from ubcs_auxiliary.save_load_object import load_from_file
    from time import time
    import timeit
    from numpy import array
    preparation = ""
    testcode = 'data = mono12packed_to_image(rawdata,height,width,mask)'

    preparation += "from ubcs_auxiliary.save_load_object import load_from_file;"
    preparation += "from lcp_video.analysis import get_mono12packed_conversion_mask;"
    preparation += "from lcp_video.analysis import mono12packed_to_image;"
    preparation += "rawdata = load_from_file('lcp_video/test_data/flir_rawdata_mono12packed.pkl');"
    preparation += "length=1,height = 2048; width = 2448;"
    preparation += "mask = get_mono12packed_conversion_mask(int(width*height*1.5));"
    preparation += 'from numpy import vstack, tile, hstack, arange,reshape,vstack, tile, hstack, arange,reshape;'

    print('tested code')
    print("l1: data_Nx8 = ((rawdata.reshape((-1,1)) & (2**arange(8))) != 0)")
    print("l2: data_N8x1 = data_Nx8.flatten()")
    print("l3: data_Mx12 = data_N8x1.reshape((int(rawdata.shape[0]/1.5),12))")
    print("l4: data = (data_Mx12*mask).sum(axis=1)")
    print("l5: data.reshape((length,height,width)).astype('int16')")

    t = timeit.Timer(testcode,preparation)
    temp = array(t.repeat(4,20))/20
    print(f"full code: {round(temp.mean(),3)} +- {round(temp.std(),3)}")



    testcode = "data_Nx8 = ((rawdata.reshape((-1,1)) & (2**arange(8))) != 0);"
    t = timeit.Timer(testcode,preparation)
    temp = array(t.repeat(4,20))/20
    print(f"line1: {round(temp.mean(),3)} +- {round(temp.std(),3)}")
    preparation +=testcode

    testcode = "data_N8x1 = data_Nx8.flatten();"
    t = timeit.Timer(testcode,preparation)
    temp = array(t.repeat(4,20))/20
    print(f"line2: {round(temp.mean(),3)} +- {round(temp.std(),3)}")
    preparation +=testcode

    testcode = "data_Mx12 = data_N8x1.reshape((int(rawdata.shape[0]/1.5),12)).astype('int16');"
    t = timeit.Timer(testcode,preparation)
    temp = array(t.repeat(4,20))/20
    print(f"line3: {round(temp.mean(),3)} +- {round(temp.std(),3)}")
    preparation +=testcode


    testcode = "data = (data_Mx12*mask).sum(axis=1);"
    t = timeit.Timer(testcode,preparation)
    temp = array(t.repeat(4,20))/20
    print(f"line4: {round(temp.mean(),3)} +- {round(temp.std(),3)}")
    preparation +=testcode

    testcode = "data.reshape((height,width)).astype('int16');"
    t = timeit.Timer(testcode,preparation)
    temp = array(t.repeat(4,20))/20
    print(f"line5: {round(temp.mean(),3)} +- {round(temp.std(),3)}")
    preparation +=testcode

def benchmark_mono12p_to_image(N = 2):
    #mono12packed_to_image(rawdata, height, width, mask)
    from ubcs_auxiliary.save_load_object import load_from_file
    from time import time
    import timeit
    from numpy import array
    preparation = ""
    testcode = 'data = mono12p_to_image(rawdata,length,height,width,mask,mask8)'

    preparation += "from ubcs_auxiliary.save_load_object import load_from_file;"
    preparation += 'from numpy import vstack, tile, hstack, arange,reshape, vstack, tile, hstack, arange,reshape, packbits, reshape, int16, concatenate, array;'
    preparation += "from lcp_video.analysis import mono12p_to_image;"
    preparation += "from lcp_video.analysis import get_mono12p_conversion_mask,get_mono12p_conversion_mask_8bit;"
    preparation += f"length = {int(N)};height = 2048; width = 2448;"
    preparation += "rawdata = load_from_file('lcp_video/test_data/flir_rawdata_mono12p.pkl'); rawdata = array([rawdata]*length);"
    preparation += "mask = get_mono12p_conversion_mask(int(width*height*length*1.5));"
    preparation += "mask8 = get_mono12p_conversion_mask_8bit(int(width*height*length*1.5));"



    t_full = timeit.Timer(testcode,preparation)

    print('tested code')
    print("1: data_Nx8 = ((rawdata.reshape((-1,1)) & (mask8)) != 0)")
    print("2: data_N8x1 = data_Nx8.flatten()")
    print("3: data_Mx12 = data_N8x1.reshape((int(rawdata.shape[-1]*length/1.5),12)).astype('int16')")
    print("4: data = (data_Mx12 * mask)).T.sum(axis=0)")
    print("5: data.reshape((length,height,width)).astype('int16')')")
    temp = array(t_full.repeat(4,20))/20
    print(f"Full function: {round(temp.mean(),3)} +- {round(temp.std(),3)}")

    testcode = 'data_Nx8 = ((rawdata.reshape((-1,1)) & (mask8)) != 0);'
    t = timeit.Timer(testcode,preparation)
    temp = array(t.repeat(4,20))/20
    print(f"line1: {round(temp.mean(),3)} +- {round(temp.std(),3)} with min of {round(temp.min(),3)} ")

    preparation +=testcode
    testcode = 'data_N8x1 = data_Nx8.flatten();'
    t = timeit.Timer(testcode,preparation)
    temp = array(t.repeat(4,20))/20
    print(f"line2: {round(temp.mean(),3)} +- {round(temp.std(),3)} with min of {round(temp.min(),3)} ")

    preparation +=testcode
    testcode = "data_Mx12 = data_N8x1.reshape((int(rawdata.shape[-1]*length/1.5),12));"
    t = timeit.Timer(testcode,preparation)
    temp = array(t.repeat(4,20))/20
    print(f"line3: {round(temp.mean(),3)} +- {round(temp.std(),3)} with min of {round(temp.min(),3)} ")


    preparation +=testcode
    testcode = "data = (data_Mx12*mask).T.sum(axis=0);"
    t = timeit.Timer(testcode,preparation)
    temp = array(t.repeat(4,20))/20
    print(f"line4: {round(temp.mean(),3)} +- {round(temp.std(),3)} with min of {round(temp.min(),3)} ")

    preparation +=testcode
    testcode = "data.reshape((length,height,width)).astype('int16');"
    t = timeit.Timer(testcode,preparation)
    temp = array(t.repeat(4,20))/20
    print(f"line5: {round(temp.mean(),3)} +- {round(temp.std(),3)} with min of {round(temp.min(),3)} ")

def profile_mono12p_to_image():
    from ubcs_auxiliary.save_load_object import load_from_file
    from lcp_video.analysis import mono12p_to_image
    mono12p = load_from_file('lcp_video/test_data/flir_rawdata_mono12p.pkl')
    height = 2048
    width = 2448
    data = mono12p_to_image(mono12p,height,width)




def benchmark_u12_to_16():
    #mono12packed_to_image(rawdata, height, width, mask)
    from ubcs_auxiliary.save_load_object import load_from_file
    from time import time
    import timeit
    from numpy import array,ones,uint8
    preparation = ""
    testcode = 'data = u12_to_16(rawdata, height=height, width = width)'

    preparation += "from ubcs_auxiliary.save_load_object import load_from_file;\n"
    preparation += 'from numpy import vstack, tile, hstack, arange,reshape, vstack, tile, hstack, arange,reshape, packbits, reshape, int16, concatenate, array, ones, uint8;\n'
    preparation += "from lcp_video.analysis import u12_to_16;\n"
    preparation += f"height = 3000; width = 4096;\n"
    preparation += "rawdata = load_from_file('lcp_video/test_data/mono12p_dataset_12Mpixels.pkl');\n "

    t_full = timeit.Timer(testcode,preparation)

    print("Preparation")
    print(preparation)
    print('tested code')
    print("    arr = rawdata.reshape(-1,3) \n        byte_even = arr[:,0]+256*(bitwise_and(arr[:,1],15)) \n \
        byte_odd = right_shift(bitwise_and(arr[:,1],240),4) + right_shift(256*arr[:,2],4) \n \
        img = empty(height*width,dtype='int16') \n \
        img[0::2] = byte_even\n \
        img[1::2] = byte_odd\n \
        return img.reshape(height,width) \n")
    temp = array(t_full.repeat(4,100))/100
    print(f"Full function: {round(temp.mean(),3)} +- {round(temp.std(),3)}, with min {round(temp.min(),3)} and max {round(temp.max(),3)}")

    from numpy import vstack, tile, hstack, arange,reshape, vstack, tile, hstack, arange,reshape, packbits, reshape, int16, concatenate, array, ones, uint8
    from ubcs_auxiliary.save_load_object import load_from_file;
    height = 2048; width = 2448;
    from lcp_video.analysis import mono12p_to_image;
    rawdata = load_from_file('lcp_video/test_data/flir_mono12p_rawdata.pkl');
    img = load_from_file('lcp_video/test_data/flir_mono12p_image.pkl');
    data = mono12p_to_image(rawdata, height=height, width = width)

    print((data.reshape(height,width) == img).any())

    from matplotlib import pyplot as plt
    plt.figure();plt.imshow(img)
    plt.figure();plt.imshow(data.reshape(height,width))

def benchmarks_multithreading():
    from numpy import random
    import threading
    from time import time
    import numpy as np

    data = random.randint(0,4096,(64,4000,4000)).astype('int32')
    result_sum = np.zeros((data.shape[1],data.shape[2]), dtype = ('int32'))

    def mt_sum(data,r1=0,r2=500,c1=0,c2=500):
        from numpy import sum
        result = sum(data[:,r1:r2,c1:c2],axis = 0)
        result_sum[r1:r2,c1:c2] = result

    t1 = time()
    working = []
    threads = []
    x = threading.Thread(target=mt_sum, args=(data,0,2000,0,2000))
    threads.append(x)
    working.append(1)
    x.start()

    x = threading.Thread(target=mt_sum, args=(data,2000,4000,0,2000))
    threads.append(x)
    working.append(1)
    x.start()

    x = threading.Thread(target=mt_sum, args=(data,0,2000,2000,4000))
    threads.append(x)
    working.append(1)
    x.start()

    x = threading.Thread(target=mt_sum, args=(data,2000,4000,2000,4000))
    threads.append(x)
    working.append(1)
    x.start()
    for index, thread in enumerate(threads):
        print("Main    : before joining thread %d.", index)
        thread.join()
        working[index] = 0
        print("Main    : thread %d done", index)

    while np.sum(working) !=0:
        sleep(0.001)
    t2 = time()
    print(t2-t1)

    data[:] = random.randint(0,4096,(64,4000,4000)).astype('int32')

    res = np.copy(result_sum*0)
    t1 = time()
    res[:] = np.sum(data, axis = 0)
    t2 = time()
    print(t2-t1)


    result_sum2 = np.copy(result_sum*0)
    data[:] = random.randint(0,4096,(64,4000,4000)).astype('int32')
    t1 = time()
    for i in range(12800):
        result_sum2[:,:] += data[i]
    t2 = time()
    print(t2-t1)

    return res, result_sum



class ParallelProcessing():
    def __init__(self, function, iter, *args, **kwargs):
        """
        This is a specialized wrapper around multiprocessing.Process object. It
        launches multiple separate processes with the same input 'function' but different argument that comes from an iterable object like tuple or list.

        Parameters
        ----------
        function :: function
            function object
        args :: list or tuple
            iterable of arguments
        kwargs :: dictionary
            keyword arguments that will be passed to the function

        Returns
        -------
        jobs :: list
            list of jobs

        Examples
        --------
        >>> mp = ParallelProcessing(function, (1,2,3))
        """
        from multiprocessing import Process
        print(args,kwargs)
        self.jobs = []
        for item in iter:
            p = Process(target=function, *args, **kwargs)
            p.start()
            self.jobs.append(p)

def function(N, **kwargs):
    """
    an example function for test purposes. It sleeps for N seconds. Prints statements when sleep starts and finishes.
    """
    from time import time, sleep

    print_text_kwarg(**kwargs)
    print_dict_kwarg(**kwargs)
    print_var_kwarg(**kwargs)

    print('Starting Process')
    t1 = time()
    print('[{} s] Start sleeping for {} seconds'.format(round(time()-t1,4),N))
    print('[{} s] Sleeping... '.format(round(time()-t1,4)))
    sleep(N)
    print('[{} s] Done sleeping for {} seconds'.format(round(time()-t1,4),N))

def print_text_kwarg(**kwargs):
    print(f"kwarg text = {kwargs['text']}")
def print_dict_kwarg(**kwargs):
    print(f"kwarg dict = {kwargs['dict']}")
def print_var_kwarg(**kwargs):
    print(f"kwarg var = {kwargs['var']}")

def test2():
    """
    simple test showing how to use
    """
    kwargs = {'text': 'my new text', 'var' : 15.64, 'dict' : {'key':'my new value'}}
    mp = ParallelProcessing(function, args = (1,2,3), kwargs = kwargs)



"""Simple database module.


Authors: Valentyn Stadnytskyi
Date created: 2019-11-21
Date last modified: 2020-01-18
Comment: Friedrich Schotte: Create directory as needed when writing, use "makedirs" instead of "mkdir"

Python Version: 2.7 and 3.7
"""
from logging import debug,info,warning,error

__version__ = "1.5.7"
from pdb import pm
from threading import RLock
lock = RLock()

from os import getcwd

class DataBase():
    def __init__(self,root,name):
        """
        initializes the connection to the database.

        Parameters
        ----------
        root :: (string)
            path or environment variable
        name :: (string)
            name of the database

        Returns
        -------

        Examples
        --------
        >>> write()
        """
        self.filename, self.root, self.name = self.get_filename(root,name)
        self.database = self.read()

    def get_filename(self,root,name):
        """
        generates filename based on input root and name. There are few special keywords like TEMP, LOCAL that will obtain the locations of temporary and local directories. The function also ensures that the path is correct in different platforms.

        Parameters
        ----------
        root :: (string)
            path or environment variable
        name :: (string)
            name of the database

        Returns
        -------
        tuple :: (filename, root, name)
            returns a tuple with 3 entries filename, root, name: (0) full filename (1) full path from the home directory; (2) the name of the file.

        Examples
        --------
        >>> filename, root, name = get_filename(root, 'name.db')
        """
        from platform import system
        from os.path import join
        name += '_db.py'
        if root == "TEMP":
            from tempfile import gettempdir
            root = join(gettempdir(),'SavedProperty')
        elif root == "LOCAL" or root == "":
            from os import getcwd
            root = join(getcwd(),'SavedProperty')
        else:
            root = join(root,'SavedProperty')
        filename = join(root,name)

        return filename, root, name

    def read(self):
        """
        reads the database file

        Parameters
        ----------

        Returns
        -------
        dictionary :: (dictionary)
            returns a dictionary of all data in the database file.

        Examples
        --------
        >>> data = read()
        """
        import yaml
        from os.path import exists
        if exists(self.filename):
            with lock:
                with open(self.filename) as handle:
                    try:
                        data = yaml.safe_load(handle.read())  # (2)
                        return data
                    except yaml.YAMLError:
                        return {}  # (3)
        else:
            return {}

    def write(self):
        """
        writes current dictionary saved in memory to the database file

        Parameters
        ----------

        Returns
        -------

        Examples
        --------
        >>> write()
        """
        # Created directory if such doesn't exist
        from os.path import dirname,exists
        from os import makedirs
        import yaml

        directory = dirname(self.filename)
        
        if not exists(directory):
            makedirs(directory)

        data = self.database
        with lock:
            with open(self.filename, 'w') as handle:
                yaml.dump(data, handle)

    def sync(self):
        """
        reads the database file and updates the self.database variable accordingly.

        Parameters
        ----------

        Returns
        -------

        Examples
        --------
        >>> sync()
        """
        self.database = self.read()


    def close(self):  # (4)
        pass

class SavedProperty():
    def __init__(self,db,name,value):
        self.name = name
        self.db = db
        self.default_value = value

    def init(self):
        """
        wrapper that returns property object with get and set functions.

        Parameters
        ----------

        Returns
        -------

        Examples
        --------
        >>> from saved_property import DataBase, SavedProperty
        >>> db = DataBase(root = '', name = 'simpledb')
        >>> d = SavedProperty(db,'d',5.0).init()
        """
        return property(self.get,self.set)

    def set(self,name,value) :
        """
        setter for SavedProperty. sets value  in the database

        Parameters
        ----------
        value :: (object)
            value saved to the database.

        Returns
        -------

        Examples
        --------
        >>>
        """
        self.db.database[self.name] = value
        self.db.write()

    def get(self,name):
        """
        getter for SavedProperty. returns saved value if it exists in the database, otherwise returns default_value

        Parameters
        ----------

        Returns
        -------
        value :: (object)
            returns a value stored with the selected key in the database

        Examples
        --------
        >>>
        """
        if self.name in list(self.db.database.keys()):
            return self.db.database[self.name]
        else:
            return self.default_value



#!/bin/env python
"""
The server LL code for the hierarchical level instrumentation

author: Valentyn Stadnytskyi NIH\LCP
data: Sept 2018 - Nov 2 2018

The server has the basic set of functions:
0) init
1) close
2) abort
3) subscribe (on/off)
4) snapshot
5) task
6)

0.0.2 - code under development
"""
__version__ = "0.0.3"
__date__ = "November 2, 2018"


from time import time, sleep, clock
import sys
import os.path
import struct
from pdb import pm
from time import gmtime, strftime, time
from logging import debug,info,warning,error
import traceback
if sys.version_info[0] ==3:
    from _thread import start_new_thread
else:
    from thread import start_new_thread

from struct import pack, unpack
from timeit import Timer, timeit

from datetime import datetime

import msgpack
import msgpack_numpy as m
import socket

import platform
server_name = platform.node()
from threading import Thread
class Client_LL(Thread):

    def __init__(self, name = ''):
        """
        to initialize an instance and create main variables
        """
        Thread.__init__(self)
        if len(name) == 0:
            self.name = 'test_communication_LL'
        else:
            self.name = name
        self.running = False
        self.network_speed = 12**6 # bytes per second
        self.push_subscribe_lst = []
        self.last_call_lst = []
        self.connection_timeout = 3.0 #timeout in seconds
        self.device = {}

    def init_server(self):
        '''
        Proper sequence of socket server initialization
        '''
        self._set_commands()
        self.sock = self.init_socket()
        if self.sock is not None:
            self.running = True
        else:
            self.running = False
        if self.running:
            self.start()

    def stop(self):
        self.running = False
        self.sock.close()

    def init_socket(self):
        '''
        initializes socket for listening, creates sock and bind to '' with a port somewhere between 2030 and 2050
        '''
        import socket
        ports = range(2030,2050)
        for port in ports:
            try:
                sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                sock.bind(('', port))
                self.port = port
                sock.listen(100)
                flag = True
            except:
                error(traceback.format_exc())
                flag = False
            if flag:
                break
            else:
                sock = None
        return sock

    def _set_commands(self):
        """
        Set type definition, the dictionary of command excepted by the server
        Standard supported commands:
        - "help"
        - "init"
        - "close"
        - "abort"
        - "snapshot"
        - "subscribe"
        - "task"
        - "controlls"
        - "indicators"
        - "push_updates"
        """
        self.commands = {}
        self.commands[b'help'] = None
        self.commands[b'init'] = None
        self.commands[b'close'] = None
        self.commands[b'abort'] = None
        self.commands[b'snapshot'] = None
        self.commands[b'subscribe'] = None
        self.commands[b'task'] = None
        self.commands[b'controls'] = None
        self.commands[b'indicators'] = None
        self.commands[b'push_updates'] = None


    def set_tasks(self):
        self.tasks = {}
        self.tasks['push updates'] = 'push updates'


    def _get_commands(self):
        """
        returns the dictionary with all supported commands
        """
        return self.commands

    def _start(self):
        '''
        creates a separete thread for server_thread function
        '''
        start_new_thread(self._run,())

    def _run(self):
        """
        runs the function _run_once in a while True loop
        """
        self.running = True
        while self.running:
            self._run_once()
        self.running = False

    def run(self):
        """
        runs the function _run_once in a while True loop
        """
        self.running = True
        while self.running:
            self._run_once()
        self.running = False

    def _run_once(self):
        """
        creates a listening socket.
        """
        client, addr = self.sock.accept()
        debug('Client has connected: %r,%r' %(client,addr))
        self._log_last_call(client, addr)
        msg_in = self._receive(client)
        try:
            pass
        except:
            msg_in = {b'command':b'unknown',b'message':b'unknown'}
            error('%r, with input message%r' %(traceback.format_exc(),msg_in))
        debug(msg_in)
        msg_out = self._receive_handler(msg_in,client)
        self._send(client,msg_out)

    def _transmit_handler(self,command = '', message = ''):
        from time import time
        res_dic  = {}
        res_dic[b'command'] = command
        res_dic[b'time'] = time()
        res_dic[b'message'] = message
        return res_dic

    def _receive_handler(self,msg_in,client):
        """
        the incoming msg_in has N mandatory fields:  command, message and time
        """
        from time import time
        res_dic = {}
        #the input msg has to be a dictionary. If not, ignore. FIXIT. I don't know how to do it in Python3
        self.last_received_message = msg_in

        try:
            keys = msg_in.keys()
            command = msg_in[b'command']
            message_in =  msg_in[b'message']
            res_dic[b'command'] = command
            flag = True
            if self.commands[command] != None:
                self.commands[command](msg_in = message_in,client = client)
            else:
                flag = False
                err = 'the command %r is not supporte by the server' % command
            if not flag:
                debug('command is not recognized')
                res_dic[b'command'] = 'unknown'
                res_dic[b'message'] = 'The quote of the day: ... . I hope you enjoyed it.'
                res_dic[b'flag'] = flag
                res_dic[b'error'] = err
            else:
                res_dic[b'flag'] = flag
                res_dic[b'error'] = ''
        except:
            error(traceback.format_exc())
            res_dic[b'command'] = 'unknown'
            res_dic[b'message'] = 'The quote of the day: ... . I hope you enjoyed it.'
            res_dic[b'flag'] = True
            res_dic[b'error'] = ''
        res_dic[b'time'] = time()

        return res_dic


    def _receive(self,client):
        """
        descritpion:
        client sends 20 bytes with a number of expected package size.
        20 bytes will encode a number up to 10**20 bytes.
        This will be enough for any possible size of the transfer

        input:
        client - socket client object

        output:
        unpacked data
        """
        import msgpack
        import msgpack_numpy as msg_m
        a = client.recv(20)
        length = int(a)
        sleep(0.01)
        if length != 0:
            msg_in = ''.encode()
            while len(msg_in) < length:
                msg_in += client.recv(length - len(msg_in))
                sleep(0.01)
        else:
            msg_in = ''
        return msgpack.unpackb(msg_in, object_hook=msg_m.decode)

    def _send(self,client,msg_out):
        """
        descrition:
        uses msgpack to serialize data and sends it to the client
        """
        debug('command send %r' % msg_out)
        msg = msgpack.packb(msg_out, default=m.encode)
        length = str(len(msg))
        if len(length)!=20:
            length = '0'*(20-len(length)) + length
        try:
            client.sendall(length.encode())
            client.sendall(msg)
            flag = True
        except:
            error(traceback.format_exc())
            flag = False
        return flag

    def _connect(self,ip_address,port):
        server = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        try:
            server.settimeout(self.connection_timeout)
            server.connect((ip_address , port))
            debug("Connection success!")
        except:
            error('%r' %(traceback.format_exc()))
            server = None
        return server

    def _transmit(self,command = '', message = '' ,ip_address = '0.0.0.0',port = 2030):
##        if isinstance(message,dict):
##            message[b'ip_address'] = ip_address
##            message[b'port'] = port
##        else:
##            message = {}
##            message[b'ip_address'] = ip_address
##            message[b'port'] = port
        msg_out = self._transmit_handler(command = command, message = message)
        #print('msg_out %r' % msg_out)
        server = self._connect(ip_address = ip_address,port = port)
        #print('server %r' % server)
        debug(server)
        if server is not None:
            flag = self._send(server,msg_out)
            #print('flag %r' % flag)

            response_arg = self._receive(server)
            #print('response_arg %r' % response_arg)

            self._client_close(server)
        else:
            response_arg = None
        return response_arg

    def _client_close(self,server):
        server.close()
        return server

    def _log_last_call(self,client,addr):
        from time import time
        self.last_call_lst = [{'time':time(),'ip_address':addr[0]}]
        #if len(self.last_call_lst) > 10:
            #self.last_call_lst.pop(0)

#***************************************************
#*** wrappers for basic response functions *********
#***************************************************

    def help(self):
        msg = {}
        msg[b'client name']= self.name
        msg[b'client port']= self.port
        msg[b'client ip address']= socket.gethostbyname(socket.gethostname())
        msg[b'client version']= __version__
        msg[b'commands'] = self.commands
        #msg['tasks'] = self.dev_get_tasks
        try:
            try:
                getattr(device.controls,'running')
            except:
                err = 'device offline'
            msg[b'device name'] = device.name
            msg[b'device version'] = device.version
            msg[b'controls'] = self.controls()
            msg[b'indicators'] = self.indicators()
        except:
            msg[b'controls'] = 'device offline'
            msg[b'indicators'] = 'device offline'
        return msg



    def identify(self, ip_address = '', port = 0, device = None):
        debug('init command request to  %r %r ' % (ip_address,port))
        if isinstance(device,dict):
            try:
                ip_address = device[b'ip_address']
                port = device[b'port']
                response = self._transmit(command = b'help',
                             message = '',
                             ip_address = ip_address,
                             port = port)
            except:
                err = traceback.format_exc()
                error(err)
                response = err
        else:
            response = self._transmit(command = b'help',
                             message = '',
                             ip_address = ip_address,
                             port = port)
        return response

    def init(self, device = {b'ip_address':'127.0.0.1',
                             b'port':2040,
                             b'name':b'localhost'}):
        debug('init command request to  %r %r ' % (device[b'ip_address'],device[b'port']))
        if isinstance(device,dict):
            try:
                ip_address = device[b'ip_address']
                port = device[b'port']
                response = self._transmit(command = b'init',
                             message = '',
                             ip_address = ip_address,
                             port = port)
            except:
                err = traceback.format_exc()
                error(err)
                response = err
        else:
            response = self._transmit(command = b'init',
                             message = '',
                             ip_address = ip_address,
                             port = port)
        return response

    def close(self, ip_address = '127.0.0.1', port = 2030,
              device = None):
        debug('close command request to  %r %r ' % (ip_address,port))
        if isinstance(device,dict):
            try:
                ip_address = device[b'ip_address']
                port = device[b'port']
            except:
                error(err)
        response = self._transmit(command = b'close',
                         message = '',
                         ip_address = ip_address,
                         port = port)
        return response

    def cstart(self,ip_address = '127.0.0.1', port = 2030,
              device = None):
        debug('close command received with message = %r' % msg)
        if isinstance(device,dict):
            try:
                ip_address = device[b'ip_address']
                port = device[b'port']
            except:
                error(err)
        response = self._transmit(command = b'start',
                     message = '',
                     ip_address = ip_address,
                     port = port)
        return response

    def stop(self,ip_address = '127.0.0.1', port = 2030,
              device = None):
        debug('close command received with message = %r' % msg)
        if isinstance(device,dict):
            try:
                ip_address = device[b'ip_address']
                port = device[b'port']
            except:
                error(err)
        response = self._transmit(command = b'stop',
                         message = '',
                         ip_address = ip_address,
                         port = port)
        return response

    def abort(self, ip_address = '', port = 0,
              device = None):
        debug('abort command request to  %r %r ' % (ip_address,port))
        if isinstance(device,dict):
            try:
                ip_address = device[b'ip_address']
                port = device[b'port']
            except:
                error(err)
        response = self._transmit(command = b'abort',
                         message = '',
                         ip_address = ip_address,
                         port = port)
        return response

    def snapshot(self, ip_address = '', port = 0, device = None):
        debug('snapshot command request to  %r %r ' % (ip_address,port))
        if isinstance(device,dict):
            try:
                ip_address = device[b'ip_address']
                port = device[b'port']
                response = self._transmit(command = b'snapshot',
                             message = '',
                             ip_address = ip_address,
                             port = port)
            except:
                err = traceback.format_exc()
                error(err)
                response = err
        else:
            response = self._transmit(command = b'snapshot',
                             message = '',
                             ip_address = ip_address,
                             port = port)
        return response

    def task(self,msg):
        if isinstance(msg,list) and isinstance(msg[0],list):
            device.set_task_queue(msg)
            response = len(server.controls({b'task_queue':None})[b'controls'][b'task_queue'][0])
        else:
            response = 'error'
        return response

    def controls(self,
                   controls = {b'all':None},
                   device = {b'ip_address':'127.0.0.1',
                             b'port':2030,
                             b'name':b'localhost'}):
        """
        return current control(controls) according to the input dictionary.
        example:
        the command->
        controls(msg = {b'controls':{'button1':None,'button2':None,'scanrate':2.0}})
        will set scanrate to 2.0 and return current setting for button1, button2 and scanrate
        """
        debug('controls command received: %r' % controls)
        controls_dic = {}

        for item in list(controls.keys()):
            controls_dic[item] = controls[item]
        if isinstance(device,dict):
            try:
                ip_address = device[b'ip_address']
                port = device[b'port']
            except:
                error(err)
        response = self._transmit(command = b'controls',
                         message = controls_dic,
                         ip_address = ip_address,
                         port = port)
        return response
    def get_control(self,control = '',device = {b'ip_address':None,b'port': None}):
        """
        gets the value of the control with the name control from device dictionary
        """
        if control != '':
            response = self.controls(device = device,
                                     controls = {control:None})[b'message'][b'controls'][control]
        else:
            response = None
        return response
    def set_control(self,control = '',
                    device = {b'ip_address':None,b'port' : None},
                    value = None):
        if control != "":
            res_dic = self.controls(device = device, controls = {control:value})
            response = res_dic[b'message'][b'controls'][control]
        else:
            print(res_dic[b'error'])
            error(res_dic[b'error'])
            response = None
        return response

    def indicators(self,
                   indicators = {b'all':None},
                   device = {b'ip_address':'127.0.0.1',
                             b'port':2030,
                             b'name':b'localhost'}):
        """
        this function takes 7us to execute probably because of hashtable,etc.
        If I call the function directly, I can execute it in 1 us
        """
        indicators_dic = {}

        for item in list(indicators.keys()):
            indicators_dic[item] = indicators[item]
        if isinstance(device,dict):
            try:
                ip_address = device[b'ip_address']
                port = device[b'port']
            except:
                err = traceback.format_exc()
                error(err)
        response = self._transmit(command = b'indicators',
                         message = indicators_dic,
                         ip_address = ip_address,
                         port = port)

        return response
    def get_indicator(self,indicator = '', device = None):
        response = self.indicators(device = device,
                                   indicators = {indicator:None})[b'message'][b'indicators'][indicator]
        return response


    def subscribe(self,ip_address ='128.231.5.229' ,port = 2030,device = {},indicators = ['all'],controls = ['all']):
        if isinstance(device,dict):
            try:
                ip_address = device[b'ip_address']
                port = device[b'port']
            except:
                error(err)
        response = self._transmit(command = b'subscribe',
                         message = {b'port':self.port, b'indicators':indicators, b'controls':controls, b'ip_address':socket.gethostbyname(socket.gethostname())},
                         ip_address = ip_address,
                         port = port)
        return response

    def buffers_update(self, ip_address ='128.231.5.229' ,port = 2030,
                        buffers = {b'all':None},
                        device = None):
        """
        this function takes 7us to execute probably because of hashtable,etc.
        If I call the function directly, I can execute it in 1 us
        """
        buffers_dic = {}
        for item in list(buffers.keys()):
            buffers_dic[item] = buffers[item]
        if isinstance(device,dict):
            try:
                ip_address = device[b'ip_address']
                port = device[b'port']
            except:
                error(err)
        response = self._transmit(command = b'buffers_update',
                         message = buffers_dic,
                         ip_address = ip_address,
                         port = port)

        return response





#######
### Test fuctions
########



client = Client_LL()

#from DAQ_dummy_LL import daq_dummy as device



if __name__ == "__main__": #for testing
    from tempfile import gettempdir
    import logging
    logging.basicConfig(filename=gettempdir()+'/XLI_client_LL.log',
                        level=logging.DEBUG, format="%(asctime)s %(levelname)s: %(message)s")
    self = client
    client.init_server()
    #server._run_once()
    device = {}

    print("device['ii-50-DL'] = {b'ip_address':'172.17.200.36',b'port':2030}")
    print("device['ii-50-SL'] = {b'ip_address':'172.17.200.36',b'port':2031}")
    print("device['ii-50-AL'] = {b'ip_address':'172.17.200.36',b'port':2032}")
    print("device['icarus-0'] = {b'ip_address':'128.231.5.84',b'port':2030}")
    print("device['icarus-0'] = {b'ip_address':'128.231.5.84',b'port':2030}")
    print("device['icarus-0'] = {b'ip_address':'128.231.5.84',b'port':2030}")


    print('Client: is running at %r,%r'
          %(client.sock.getsockname(),socket.gethostbyname(socket.gethostname())))
    print("client._transmit(command = 'init', message = '', ip_address = '%s', port = %r)"
          % ('127.0.0.1',2035))
    print("client.indicators(device = device['icarus-0'], indicators = {b'running':None})")
    print("client.subscribe(ip_address = '127.0.0.1', port = 2035, device  = None)")
    print("data = client._transmit(command = b'read_queue', message = {b'number':70000},ip_address = '127.0.0.1', port = 2030)[b'message']")



"""
XLevel delocalized instrumentation class
author: Valentyn Stadnytskyi
Created: November 2017
Last modified: March 7 2019
"""

__version__ = '0.0.2'

import traceback
import psutil, os
import platform #https://stackoverflow.com/questions/110362/how-can-i-find-the-current-os-in-python
p = psutil.Process(os.getpid()) #source: https://psutil.readthedocs.io/en/release-2.2.1/


from numpy import nan, mean, std, nanstd, asfarray, asarray, hstack, array, concatenate, delete, round, vstack, hstack, zeros, transpose, split, unique, nonzero, take, savetxt, min, max
from serial import Serial
from time import time, sleep
import sys
import os.path
import struct
from pdb import pm
from time import gmtime, strftime, time
from logging import debug,info,warning,error


from struct import pack, unpack
from timeit import Timer, timeit

from threading import Thread, Event, Timer, Condition
from datetime import datetime

class GUITemlpate():
    def __init__(self):
        pass

    def initGUI(self):
        pass

class IndicatorsTemplate():

    def __init__(self, object):
        self.object = object
        self.list = []

    def keys(self):
        if len(self.list) == 0:
            self.list = list(self.get().keys())
        return self.list

    def get(self, value = None):
        """
        returns a dictionary with all indicators. Every entry needs to
        be added manually. This helps to orginize what is avaiable to the outsie
        requestor.
        """
        dic = {}
        dic[b'running'] = self.running

        return dic



    def get_running(self):
        """
        default get_running function in the intrumentation library
        """
        raise NotImplementedError
        #return response
    def set_running(self,value):
        """
        default set_running function in the intrumentation library
        """
        raise NotImplementedError
    running = property(get_running,set_running)

class ControlsTemplate():
    def __init__(self, object):
        self.list = []
        self.object = object

    def keys(self):
        if len(self.list) == 0:
            self.list = list(self.get().keys())
        return self.list


    def get(self):
        dic = {}
        dic[b'variable'] = 0
        return dic

    def set(self,new_controls = [{b'temp':False}]):
        for key in list(new_controls.keys()):
            setattr(self,key.decode("utf-8"),new_controls[key])
        response = self.get()
        return response

    def get_variable(self):
        try:
            response = getattr(self.object,'variable')
        except:
            response = None #device.controls.running
            warning(traceback.format_exc())
        return response
    def set_variable(self,value):
        """
        indicators cannot be set from outside. This method is used to set
        an indicator from other instances within this proccess.
        """
        try:
            setattr(self.object,'variable',value)
        except:
            error(traceback.format_exc())
    variable = property(get_variable,set_variable)

    def get_task_queue(self):
        """
        """
        raise NotImplementedError

    def set_task_queue(self,value):
        """
        """
        raise NotImplementedError
    variable = property(get_variable,set_variable)




class XLevelTemplate():

    #inds = IndicatorsTemplate(self)
    #ctrls = ControlsTemplate(self)
    """circular buffers dictionary contains information about all circular buffers and their type (Server, Client or Queue)"""
    circular_buffers = {}

    def __init__(self):
        #Thread.__init__(self)
        self.running = False
        #self.daemon = False # OK for main thread to exit even if instance is still running
        self.description = ''

    def first_time_setup(self):
        """aborts current operation
        """
        raise NotImplementedError


    def init(self, msg_in = None, client = None):
        """
        does proper start of the XLevel code and can be called remoutely
        """
        raise NotImplementedError
        response = {}
        response[b'flag'] = flag
        response[b'message'] = message
        response[b'error'] = err
        return response

    def abort(self,msg_in = None, client = None):
        """aborts current operation
        """
        raise NotImplementedError
        response = {}
        response[b'flag'] = flag
        response[b'message'] = message
        response[b'error'] = err
        return response

    def close(self,msg_in = None, client = None):
        """aborts and completely stops the XLevel code
        """
        raise NotImplementedError

        response = {}
        response[b'flag'] = flag
        response[b'message'] = message
        response[b'error'] = err
        return response



    def help(self,msg_in = None, client = None):
        """
        returns help information about this XLevel program

        ####
        EXAMPLE:
        response = {}
        response[b'name'] = self.name
        response[b'controls'] = self.ctrls.get()
        response[b'indicators'] = self.inds.get()
        return response
        """
        raise NotImplementedError

        response = {}
        response[b'flag'] = flag
        response[b'message'] = message
        response[b'error'] = err
        return response

    def snapshot(self,msg_in = None, client = None):
        """
        returns snapshot of selected or all controls and indicators.
        see example below

        EXAMPLE:
        response = {}
        response[b'name'] = self.name
        response[b'controls'] = self.ctrls.get()
        response[b'indicators'] = self.inds.get()
        return response
        """
        raise NotImplementedError

        response = {}
        response[b'flag'] = flag
        response[b'message'] = message
        response[b'error'] = err
        return response



    def controls(self,msg_in = {b'all':None}, client = None):
        """
        return current control(controls) according to the input dictionary.
        example:
        the command->
        controls(msg = {b'controls':{'button1':None,'button2':None,'scanrate':2.0}})
        will set scanrate to 2.0 and return current setting for button1, button2 and scanrate
        """
        debug('icarusSL controls command received: %r' % msg_in)
        response = {}
        msg_out = {}
        msg_out[b'message'] = {}
        err = ''
        flag = True
        if isinstance(msg_in,dict):
            if b'all' in list(msg_in.keys()):
                    response = self.ctrls.get()
                    ctrls_keys = self.ctrls.keys()
            else:
                ctrls_keys = self.ctrls.keys()
                for key in msg_in:
                    if msg_in[key] is None:
                        if key in ctrls_keys:
                            response[key] = getattr(self.ctrls,key.decode("utf-8"))
                        else:
                            response[key] = None
                            err = "control doesn't exist"
                    else:
                        if key in ctrls_keys:
                            setattr(self.ctrls,key.decode("utf-8"),msg_in[key])
                            response[key] = getattr(self.ctrls,key.decode("utf-8")) #msg_in[key]
                        else:
                            response[key] = None
                            err = "control doesn't exist"

        msg_out[b'flag'] = flag
        msg_out[b'message'][b'controls'] = response
        msg_out[b'error'] = err
        return msg_out

    def indicators(self, msg_in = {b'all':None}, client = None):
        """
        this function takes 7us to execute probably because of hashtable,etc.
        If I call the function directly, I can execute it in 1 us
        """
        debug('icarusSL indicators command received: %r' % msg_in)
        response = {}
        msg_out = {}
        msg_out[b'message'] = {}
        err = ''
        flag = True
        if isinstance(msg_in,dict):
            if b'all' in list(msg_in.keys()):
                    response = self.inds.get()
                    inds_keys = self.inds.keys()
            else:
                inds_keys = self.inds.keys()
                for key in msg_in:
                    if msg_in[key] is None:

                        if key in inds_keys:
                            response[key] = getattr(self.inds,key.decode("utf-8"))
                        else:
                            response[key] = None
                            err = "indicator doesn't exist"
                    else:
                        if key in inds_keys:
                            setattr(self.inds,key.decode("utf-8"),msg_in[key])
                            response[key] = getattr(self.inds,key.decode("utf-8"))
                        else:
                            response[key] = None
                            err = 'error'
        msg_out[b'flag'] = flag
        msg_out[b'message'][b'indicators'] = response
        msg_out[b'error'] = err
        return msg_out

    def notify_subscribers(self, msg_in = None, client = None):

        raise NotImplementedError

    def schedule(self, task_list = []):
        try:
            self.controls.set({b'task_queue':task_list})
            flag = True
        except:
            error(traceback.format_exc())
            flag = False

        response = {}
        response[b'flag'] = flag
        response[b'message'] = message
        response[b'error'] = err
        return response
###
    ##########################################
    ### TASK SECTION: schedule, execute and abort
    #######################################
    """
    This section is dedicated to functions that take care of
    the queue task consept and help to execute it.
    """
    def schedule_task_queue(self, task_list = []):
        try:
            self.controls.set({b'task_queue':task_list})
            response = True
        except:
            error(traceback.format_exc())
            response = False
        return response

    def abort_task_queue(self):
        try:
            self.controls.set({b'task_queue':[]})
            response = True
        except:
            error(traceback.format_exc())
            response = False
        return response

    def execute_task_queue(self):
        if len(self.controls.task_queue) > 0:
            if self.controls.task_queue[0][0] - time() <=0:
                try:
                    info('executing: %r, with flag %r' %
                         (self.controls.task_queue[0],
                          self.controls.task_queue[0][1] == b'change_offset'))
                    if self.controls.task_queue[0][1] == b'change_offset' or self.controls.task_queue[0][1] == 'change_offset':

                        self.change_offset(self.controls.task_queue[0][2][b'offset'])
                    self.controls.task_queue.pop(0)
                except:
                    pass#error(traceback.format_exc())
                flag = True
            else:
                flag = False
        else:
            flag = False
        return flag

    #########
    ### OTHER TASKS GO HERE. The XLevel can support multiple additional tasks
    ### which are more complicated than just return indicators or controls
    ########
    #def start(self): This is not needed for the class Thread
    #    """"""
    #    raise NotImplementedError

    def stop(self):
        """"""
        raise NotImplementedError

    def run_once(self):
        """"""
        raise NotImplementedError

    def run(self):
        """"""
        self.running = True
        while self.running:
            self.run_once()
            sleep(1)
        raise NotImplementedError


    def get_circular_buffer(self, msg_in = None, client = None):
        """
        returns data from server circular buffer. Input, name of the circular buffer and global pointer.
        msg_in has to be a dictionary with circular buffer name
        the global_pointer sppecifies
        msg_in = {b'buffer_name':b'name',b'g_pointer':999}
        """
        msg_out = {b'message':{}}
        flag = True
        err = ''

        in_buffer_name = msg_in[b'buffer_name']
        in_g_pointer = msg_in[b'g_pointer']
        if in_buffer_name in self.circular_buffers.keys():
            g_pointer = self.circular_buffers[in_buffer_name].g_pointer
            pointer = self.circular_buffers[in_buffer_name].pointer
            size = self.circular_buffers[in_buffer_name].size
            if in_g_pointer > g_pointer:
                data = self.circular_buffers[in_buffer_name].get_all()
            elif in_g_pointer < g_pointer:
                if g_pointer - in_g_pointer > size[1]:
                    data = self.circular_buffers[in_buffer_name].get_all()
                else:
                    N = g_pointer - in_g_pointer
                    data = self.circular_buffers[in_buffer_name].get_N(N, M = pointer)
            else:
                data = None

            self.circular_buffers[in_buffer_name]

        msg_out[b'flag'] = flag
        msg_out[b'message'][b'data'] = data
        msg_out[b'message'][b'g_pointer'] = g_pointer
        msg_out[b'message'][b'pointer'] = pointer
        msg_out[b'error'] = err
        return msg_out

    def retrieve_values(self, msg_in = None , N = 2, order = 2, test_flag = None, client = None):
        """
        msg_in = {b'buffer_name':b'name',b'time_vector' = asarray([ 1,  1,  1,  1,  1,  1,  1])}
        N - number of points to concider when obtain estimated value
        """
        from numpy import argmin, argwhere, loadtxt, nanmax,\
             nanmin, nan, nanargmin, nanargmax

        from XLI.auxiliary import sort_vector, expand_vector, get_estimate


        msg_out = {b'message':{}}
        flag = True
        err = ''
        out_array = None

        if isinstance(msg_in,dict):
            in_buffer_name = msg_in[b'buffer_name']
            in_vector = sort_vector(msg_in[b'time_vector'])
            if in_buffer_name in self.circular_buffers.keys():
                ndim = self.circular_buffers[in_buffer_name].size[0]
                out_array = expand_vector(in_vector, ndim = ndim)
                for i in range(len(in_vector)):
                    for j in range(1,ndim):
                        x = self.circular_buffers[in_buffer_name].get_all()[0,:]
                        y = self.circular_buffers[in_buffer_name].get_all()[j,:]
                        x_est = in_vector[i]
                        if x_est > nanmax(x):
                            y_est = nan
                            err += 'requested value(%r) outside of the buffer boundaries. value > nanmax (%r)' % (x_est,nanmax(x))
                        elif x_est < nanmin(x):
                            y_est = nan
                            err += 'requested value(%r) outside of the buffer boundaries. value < nanmin (%r)' % (x_est,nanmin(x))
                        else:
                            idx = nanargmin((x - x_est)**2)
                            debug('idx = %r' %idx)
                            if idx >= len(x)-2:
                                N_after = 0
                                N_before = N
                            elif idx <= 1:
                                N_after = N+1
                                N_before = 0
                            else:
                                N_after = N+1
                                N_before = N
                            debug('N_after %r, N_before %r' %(N_after,N_before))

                            y_est = get_estimate(x[idx-N_before:idx+N_after],y[idx-N_before:idx+N_after],x_est, order = order)

                        out_array[j,i] = y_est

            else:
                err = "buffer doesn't exist"
                out_vector = msg_in[b'time_vector']
        msg_out[b'flag'] = flag
        msg_out[b'message'][b'out_array'] = out_array
        msg_out[b'error'] = err
        return msg_out
########################################
###  Threading Section   ###
####################################

###BINDING OF A SERVER MODULE WITH XLevel EXAMPLE

##from server_LL import server
##server.init_server(name = 'icarus-server')
##server.commands[b'init'] = icarus_SL.init
###server.commands[b'help'] = device.help
###server.commands[b'snapshot'] = device.snapshot
###server.commands[b'close'] = device.close
##server.commands[b'controls'] = icarus_SL.controls
##server.commands[b'indicators'] = icarus_SL.indicators
###server.commands[b'retrieve_values'] = device.retrieve_values
##server.commands[b'subscribe'] = server.subscribe
###server.commands[b'buffers_update'] = device.buffers_update
###server.commands[b'dump_buffers'] = icarus_SL.subscribe



if __name__ == "__main__": #for testing
    from tempfile import gettempdir
    import logging
    #logging.basicConfig(#filename=gettempdir()+'/icarus_SL.log',
     #                   level=logging.INFO, format="%(asctime)s %(levelname)s: %(message)s")



from numpy import zeros, array_equal

from ubcs_auxiliary import precision_sleep, interupt_sleep
from time import time
import random

# def test_precision_sleep():
#     """ runs multiple tests with fixed maximum allowed error (1 ms) """
#     precision = 10.0e-4
#     for i in range(100):
#         sleep_t = random.randint(1,20)/1000.0
#         t1 = time(); precision_sleep(sleep_t); t2 = time(); dt = t2-t1
#         # run tests
#         assert ((dt - sleep_t) < precision)



#!/bin/env python
# -*- coding: utf-8 -*-
"""test Queue
    by Valentyn Stadnytskyi
     created: August 2, 2019
    This is a test library to evaluate the performance of the code.
    Queue is an abstract data structure, somewhat similar to Stacks.
    Unlike stacks, a queue is open at both its ends.
    One end is always used to insert data (enqueue)
    and the other is used to remove data (dequeue)

    to run unittest: python3 -m unittest test_queue
"""

import unittest
import logging
# from numpy.testing import assert_, assert_almost_equal, assert_equal

from .. import numerical

class ImageTest(unittest.TestCase):

    def test_distance_matrix(self):
        """
        various build-in attributes.
        """
        import numpy as np
        row = np.asarray([100,200,400])
        col = np.asarray([100,200,300])
        dist = numerical.distance_matrix(row = row,col = col)
        result = np.array([[  0.        , 141.42135624, 360.55512755],
           [141.42135624,   0.        , 223.60679775],
           [360.55512755, 223.60679775,   0.        ]])
        self.assertEqual((np.around(dist,3)==np.around(result,3)).all(),True)

    def test_nearest_neibhour(self):
        import numpy as np
        row = np.asarray([100,200,400])
        col = np.asarray([100,200,300])
        nn = numerical.nearest_neibhour(row = row,col = col)
        result = np.array([[0, 1, 2],
           [1, 0, 2],
           [2, 1, 0]])
        self.assertEqual((np.around(nn,3)==np.around(result,3)).all(),True)



"""
plotting functions
author: Valentyn Stadnytskyi
data: 2017 - Nov 17 2018

"""
__vesrion__ = '1.0.0'
from time import  time, sleep
import sys
if sys.version_info[0] == 2:
    from time import clock as perf_counter
else:
    from time import perf_counter
import platform
from numpy import asarray

### Mathematical Functions
def exponential_1(x,x0,A,tau, offset):
    """
    exponential function with one exponent
    """
    from numpy import exp
    func = A*exp(-(x-x0)/tau)+offset
    return func

def linear(x,a,b):
    """
    linear function
    """
    return a + b*x

def binary_to_array(value = 0, length = 8):
    """
    takes an integer and converts it to 8 bit representation as an array.
    If float number is passed, it will be converted to int.
    """
    from numpy import arange,ndarray,nan
    value = int(value)
    binary = format(value, '#0'+str(length+2)+'b')
    arr = arange(length-1)
    for i in range(length-1):
        arr[i] = binary[length+1-i]
    return arr

def array_to_binary(arr = asarray([ 1,  1,  1,  1,  1,  1,  1])):
    """
    takes an integer and converts it to 8 bit representation as an array.
    If float number is passed, it will be converted to int.
    """
    from numpy import arange,ndarray,nan
    integer = 0
    for i in range(len(arr)):
        integer += int(arr[i]*2**(i))
    return integer

def bin_data(data  = None, x = None, axis = 1, num_of_bins = 300, dtype = 'float'):
        """returns a vector of integers on logarithmic scale starting from decade start, ending decade end with M per decade
        Parameters
        ----------
        data (numpy array)
        x (numpy array)
        axis (integer)
        num_of_bins (integer)
        dtype (string)

        Returns
        -------
        dictionary with keys: 'x',y_min','y_max''y_mean'

        Examples
        --------
        >>> from numpy import random, arange
        >>> data = random.rand(4,1000)+ 1
        >>> x_in = arange(0,data.shape[0]+1,1)
        >>> binned_data = bin_data(data  = None, x = None, axis = 1, num_of_bins = 300, dtype = 'float')

        .. plot:: ./examples_py/numerical_bin_data.py
           :include-source:

        """
        from numpy import zeros, nan,arange, nanmax, nanmin, random,nanmean, mean
        import math
        if len(data.shape) > 1:
            length = data.shape[0]
            width = data.shape[1]
        else:
            length = data.shape[0]
            width = 1

        if x is None:
            x = arange(0,length,1)

        if length <= num_of_bins:
            y_max = data
            y_min = data
            y_mean = data
            x_out = x
        else:
            y_min = zeros(shape = (width,num_of_bins), dtype = dtype)
            y_max = zeros(shape = (width,num_of_bins), dtype = dtype)
            y_mean = zeros(shape = (width,num_of_bins), dtype = dtype)
            x_out = zeros(shape = (num_of_bins,), dtype = dtype)

            for j in range(width):
                idx = 0
                for i in range(num_of_bins):
                    step = int(math.ceil(1.0*(length - idx)/(num_of_bins-i)))

                    start = idx
                    end = idx + step
                    if 'int' in dtype:
                        y_max[j,i] = int(nanmax(data[start:end,j]))
                        y_mean[j,i] = int(nanmean(data[start:end,j]))
                        y_min[j,i] = int(nanmin(data[start:end,j]))
                    else:
                        y_max[j,i] = nanmax(data[start:end,j])
                        y_mean[j,i] = nanmean(data[start:end,j])
                        y_min[j,i] = nanmin(data[start:end,j])
                    x_out[i] = mean(x[start:end])
                    idx += step
        dic = {}
        dic['x'] = x_out
        dic['y_min'] = y_min
        dic['y_max'] = y_max
        dic['y_mean'] = y_mean
        return dic

def bin_xy(y  = None, x = None, axis = 1, num_of_bins = 300, dtype = 'float'):
        """returns a vector of integers on logarithmic scale starting from decade start, ending decade end with M per decade
        Parameters
        ----------
        data (numpy array)
        x (numpy array)
        axis (integer)
        num_of_bins (integer)
        dtype (string)

        Returns
        -------
        dictionary with keys: 'x',y_min','y_max''y_mean'

        Examples
        --------
        >>> from numpy import random, arange
        >>> data = random.rand(4,1000)+ 1
        >>> x_in = arange(0,data.shape[0]+1,1)
        >>> binned_data = bin_data(data  = None, x = None, axis = 1, num_of_bins = 300, dtype = 'float')

        .. plot:: ./examples_py/numerical_bin_data.py
           :include-source:

        """
        from numpy import zeros, nan,arange, nanmax, nanmin, random,nanmean, mean
        import math

        length = y.shape[0]


        if length <= num_of_bins:
            y_max = data
            y_min = data
            y_mean = data
            x_out = x
        else:
            y_min = zeros(shape = (num_of_bins,), dtype = dtype)
            y_max = zeros(shape = (num_of_bins,), dtype = dtype)
            y_mean = zeros(shape = (num_of_bins,), dtype = dtype)
            x_out = zeros(shape = (num_of_bins,), dtype = dtype)

            idx = 0
            for i in range(num_of_bins):
                step = int(math.ceil(1.0*(length - idx)/(num_of_bins-i)))
                start = idx
                end = idx + step
                if 'int' in dtype:
                    y_max[i] = int(nanmax(y[start:end]))
                    y_mean[i] = int(nanmean(y[start:end]))
                    y_min[i] = int(nanmin(y[start:end]))
                else:
                    y_max[i] = nanmax(y[start:end])
                    y_mean[i] = nanmean(y[start:end])
                    y_min[i] = nanmin(y[start:end])
                x_out[i] = mean(x[start:end])
                idx += step
        dic = {}
        dic['x'] = x_out
        dic['y_min'] = y_min
        dic['y_max'] = y_max
        dic['y_mean'] = y_mean
        return dic

def sort_vector(in_vector = asarray([ 1,  1,  1,  1,  1,  1,  1])):
    """sorts time vector"""
    from numpy import sort
    if in_vector.ndim == 1:
        out_vector = sort(in_vector)
    return out_vector

def expand_vector(in_vector = asarray([ 1,  1,  1,  1,  1,  1,  1]),ndim = 2):
    """ makes input 1D vector as 2D with first dimenstion to be ndim"""
    from numpy import zeros,expand_dims,concatenate
    out_vector = expand_dims(in_vector, axis = 0)
    add_vector = out_vector*0
    for i in range(ndim-1):
        out_vector = concatenate((out_vector,add_vector),axis=0)
        return out_vector

def get_estimate(x,y,x_est, order = 2):
        """
        returns estimated y_est value for give x_est from real x,y data set.
        """
        from numpy import polyfit, poly1d, nanargmin
        idx = nanargmin((x - x_est)**2)
        x0 = x[idx]
        debug('x = %r' %x)
        debug('y = %r' %y)
        debug('x_est = %r' %y)
        debug('idx = %r' %idx)
        debug('x0 = %r' %x0)
        if len(x) !=0 or len(y) != 0:
            fit = polyfit(x-x0,y,order)
            ynew_fit = poly1d(fit)(x_est-x0)
        else:
            ynew_fit = nan
        return ynew_fit

def log_scale(N = 8, start = -9, end = 3, round_to = 3):
    """creates an array of numbers on logarithmic scale with:
        - number per decade
        - start decade
        - end decade
        - round_to number of digits after decimal, default is 3
        """
    from numpy import logspace, around
    arr = around(logspace(start,  end, (end-start)*N+1, endpoint=True),abs(start-round_to))
    return arr

def local_log_scale(start_dec, end_dec, N_per_dec, dtype = 'int64'):
    """returns a vector of integers on logarithmic scale starting from decade start, ending decade end with M per decade
    Parameters
    ----------
    start_dec (integer)
    end_dec (integer)
    N_per_dec (integer)
    dtype (string)

    Returns
    -------
    array (numpy array)

    Examples
    --------
    >>> arr = local_log_scale(start_dec = -9, end_dec = 1, N_per_dec = 4, dtype = 'int64')
    """
    from numpy import logspace, around
    arr = logspace(start_dec,  end_dec, (end_dec-start_dec)*N_per_dec, endpoint=False,dtype = dtype)
    return arr

def bin_on_logscale(x,y,N = 100, dN = 1, x0 = 0, M = 16, order = 1, mode = 'polyfit'):
    """
    purpose: binning of data: first N points starting from x0 are
    binned in bins of size dN and the rest is binned on logarithmic scale with M per decade

    Parameters
    ----------
    x (integer) - x-axis of data
    y (integer) - y-axis of data
    x0 (float) - the zero on x-axis
    N (int) - number of points after x0 that are binned , dN bi size, on linear scale
    dN (integer) - size of the bin for the linear scale, first N points.
    M (integer)- number of points per decade for the rest of the data

    Returns
    -------
    (y_mean, y_std, x_out, num)
    y_mean array (numpy array)
    y_std  array (numpy array)
    x_out  array (numpy array)
    num array (numpy array)

    Examples
    --------
    >>> arr = local_log_scale(start_dec = -9, end_dec = 1, N_per_dec = 4, dtype = 'int64')
    """
    if M >128:
        raise ValueError('Value of M = %r has exceeded allowed (%r)' %(M,128))

    def lin_func(x_l,x_r,y,x0, order = 2, fit_mode = 'mean'):
        """
        """
        from numpy import polyfit, poly1d, arange, std
        if y.shape[0] > 1:
            x = arange(x_l,x_r,1)
            if x.shape[0] >2:
                if fit_mode == 'polyfit':
                    #print('1',fit_mode)
                    y_fit = poly1d(polyfit(x,y,deg = order))
                    y_fit_x0 = y_fit(x0)
                    y_std = (sum((y_fit(x) - y)**2)/x.shape[0])**0.5
                elif fit_mode == 'bevington':
                    #print('2',fit_mode)
                    a,b,y_std = linear_fit(x,y)
                    y_fit_x0 = linear(x0,a,b)
                else:
                    #print('3',fit_mode)
                    y_fit_x0,y_std = mean(y), std(y)
            else:
                if fit_mode == 'polyfit':
                    y_fit = poly1d(polyfit(x,y,deg = 1))
                    y_fit_x0 = y_fit(x0)
                    y_std = (sum((y_fit(x) - y)**2)/x.shape[0])**0.5
                elif fit_mode == 'bevington':
                    a,b,y_std = linear_fit(x,y)
                    y_fit_x0 = linear(x0,a,b)
                else:
                    y_fit_x0,y_std = mean(y), std(y)


            return y_fit_x0, y_std
        else:
            return y, 0
    t={}
    k = 0
    t0 = time()

    from numpy import mean, std, arange, argwhere, concatenate,polyfit, poly1d, mod, log10
    y = transpose(y)
    x_len = x.shape[0]
    y_len = y.shape[0]
    num = x_len
    t[k] = time() - t0;k +=1;
    lin_left = arange(0,(N-1)*dN,dN)
    lin_right = arange(dN,N*dN,dN)
    lin_middle = lin_left/2.0+(lin_right-1)/2.0
    #print('lin_left.shape[0] = %r. lin_right.shape[0] = %r' %(lin_left.shape[0], lin_right.shape[0]))
    t[k] = time() - t0;k +=1;
    up_to = int(log10(num))+1
    x_log = local_log_scale(1,up_to,2*M)
    arr1  = x_log[argwhere((x_log >= lin_right[-1]))][:,0]
    arr2 = arr1[argwhere(arr1 <= x_len)][:,0]
    log_left = arr2[asarray(range(0,arr2.shape[0],2))]
    log_middle = arr2[asarray(range(1,arr2.shape[0],2))]
    log_right = arr2[asarray(range(2,arr2.shape[0],2))]
    t[k] = time() - t0;k +=1;
    right = concatenate((lin_right,log_right))
    left = concatenate((lin_left,log_left))[:len(right)]
    middle = concatenate((lin_middle,log_middle))[:len(right)]

    t[k] = time() - t0;k +=1;
    x_out = right*0.0
    y_mean = right*0.0
    y_std = right*0.0
    num = right*0.0
    #print(bogus)
    t[k] = time() - t0;k +=1;
    for i in range(x_out.shape[0]):
        #if mod(N,2):
        x_out[i],x_out_std = lin_func(left[i],right[i],x[left[i]:right[i]],middle[i], order = order, fit_mode = 'polyfit')
        #else:
           # try: x_out[i] = x[int(middle[i])]
           # except: print('bla');x_out[i] = x[int(middle[i])-1]
        y_mean[i],y_std[i] = lin_func(left[i],right[i],y[left[i]:right[i]],middle[i], order = order, fit_mode = mode)
        num[i] = x[left[i]:right[i]].shape[0]
    t[k] = time() - t0;k +=1;
    return y_mean, y_std, x_out, num


def linear_fit(x,y):
    """
    return linear fit by calcualating
    y_fit = a + b*x
    page 104 Data reduction and error analysis for the physicxal sciences Philip R. Bevington

    Parameters
    ----------
    x (1d numpy array)
    y (1d numpy array)

    Returns
    -------
    a
    b
    sigma

    Examples
    --------
    >>> a, b , sigma = linear_fit(x,y)
    """

    from numpy import isnan,nan, sum

    Sx = sum(x*1.0) #Sx_i = Sx_i-1 +x_i
    Sx2 = sum(x**2.00) #Sx2_i = Sx2_i-1 + x_i**2
    Sy = sum(y*1.0) #Sy_i = Sy_i-1 + y_i
    Sy2 = sum(y**2.0) #Sy2_i = Sy2_i-1 + y_i**2
    Sxy = sum(x*y*1.0) #Sxy_i = Sxy_i-1 + x_i*y_i
    N = x.shape[0]#N_i = N_i-1 + 1.0
    if N >= 2:
        Delta = N*Sx2 - Sx**2 # Delta_i = N_i*Sx2_i - Sx_i**2
        a = (1.0/Delta)*(Sx2*Sy-Sx*Sxy)
        b = (1.0/Delta)*(N*Sxy-Sx*Sy)
    else:
        a = None
        b = None
        #page 115
    if N > 2:
        Sigma = (1/(N-2))*(Sy2+N*a**2+(b**2)*Sx2-2*a*Sy-2*b*Sxy+2*a*b*Sx)
    else:
        Sigma = None

    return a, b, Sigma

def weighted_linear_fit(x,y,w):
    """
    return linear fit by calcualating
    y_fit = a + b*x
    page 104 Data reduction and error analysis for the physicxal sciences Philip R. Bevington

    Parameters
    ----------
    x (1d numpy array)
    y (1d numpy array)
    w (1d numpy array)

    Returns
    -------
    a
    b
    sigma_a
    sigma_b

    Examples
    --------
    >>> a, b , sigma_a, sigma_v = weighted_linear_fit(x,y,w)
    """

    from numpy import isnan,nan, sum, sqrt, min, max
    coeff = abs(x.min()-x.max())
    x = x/coeff
    Sx = sum(w*x*1.0) #Sx_i = Sx_i-1 +x_i
    Sx2 = sum(w*x**2.0) #Sx2_i = Sx2_i-1 + x_i**2
    Sy = sum(w*y*1.0) #Sy_i = Sy_i-1 + y_i
    Sy2 = sum(w*y**2.0) #Sy2_i = Sy2_i-1 + y_i**2
    Sxy = sum(w*x*y*1.0) #Sxy_i = Sxy_i-1 + x_i*y_i
    Sw = sum(w)
    N = x.shape[0]#N_i = N_i-1 + 1.0
    if N >= 2:
        Delta = (Sw*Sx2 - Sx**2) # Delta_i = N_i*Sx2_i - Sx_i**2
        a = (1.0/Delta)*(Sx2*Sy-Sx*Sxy)*coeff**2
        b = (1.0/Delta)*(Sw*Sxy-Sx*Sy)*coeff
    else:
        a = None
        b = None
        #page 115
    if N > 2:
        sigma_a = sqrt((1/Delta)*Sx2)
        sigma_b = sqrt((1/Delta)*Sw/coeff**2)
    else:
        sigma_a = sigma_b = None
    print('Delta = ',Delta)
    print('Sw2 = ',Sw)
    print('Sx = ',Sx)
    print('Sx2 = ',Sx2)
    print('coeff = ',coeff)
    print('sigma_a = ',sigma_a)
    print('sigma_b = ',sigma_b)

    return a, b, sigma_a, sigma_b

def interpolate(x,y,x_new,w = None,s = None,):
    from scipy.interpolate import UnivariateSpline
    spl = UnivariateSpline(x, y)
    y_new = spl(x_new)

def weighted_linear_fit_test_data():
    from numpy import arange, array,sqrt
    x = 5 + arange(0,150,15)/coeff
    y = array([106,80,98,75,74,73,49,38,37,22])
    w = 1/y
    return x,y,w

def weighted_linear_fit_test():
    from numpy import sqrt
    x,y,w = weighted_linear_fit_test_data()
    a,b,sigma_a,sigma_b = weighted_linear_fit(x,y,w)
    plt.figure()

    plt.errorbar(x,y,1/sqrt(w),marker='s')
    plt.plot(x,a+b*x)
    string = str(round(a,3))+','+str(round(b,3))+','+str(round(sigma_a,3))+','+str(round(sigma_b,3))
    plt.errorbar(x0,a+b*(x0),sqrt(sigma_a**2+(x0*sigma_b)**2),marker = 's')
    string += ','+str(round(sqrt(sigma_a**2+(x0*sigma_b)**2),3))
    plt.title(string)
    plt.show()
    return [sqrt(sigma_a**2+(x0*sigma_b)**2),x.max()-x.min()]

def grow_mask(mask, count=1):
    """
    Expands area where pixels have value=1 by 'count' pixels in each
    direction, including along the diagonal.

    Example: If count is 1 or omitted a single pixel grows to nine pixels.

    Parameters
    ----------
    mask (2d numpy array)
    count (integer)

    Returns
    -------
    mask (2d numpy array)

    Examples
    --------
    >>>  mask2 = grow_mask(mask,2)
    """
    from numpy import array,zeros

    if count < 1: return mask
    if count > 1: mask = grow_mask(mask,count-1)
    w,h = mask.shape
    mask2 = zeros((w,h),mask.dtype)
    mask2 |= mask
    mask2[0:w,0:h-1] |= mask[0:w,1:h] # move up by 1 pixel
    mask2[0:w,1:h] |= mask[0:w,0:h-1] # move down by 1 pixel
    mask2[0:w-1,0:h] |= mask[1:w,0:h] # move to the left by 1 pixel
    mask2[1:w,0:h] |= mask[0:w-1,0:h] # move to the right by 1 pixel

    mask2[0:w-1,0:h-1] |= mask[1:w,1:h] # move left and up by 1 pixel
    mask2[0:w-1,1:h] |= mask[1:w,0:h-1] # move left and down by 1 pixel
    mask2[1:w,0:h-1] |= mask[0:w-1,1:h] # move up and up by 1 pixel
    mask2[1:w,1:h] |= mask[0:w-1,0:h-1] # move up and down by 1 pixel

    return mask2

def enumerate_mask(mask, value = 1):
    """
    Takes a boolean mask, enumerates spots and returns enumerated mask where the intensity of a pixel indicates the spot number.
    """
    from skimage import measure
    emask = measure.label(mask==value)
    return emask

def get_array_piece(arr, center = (0,0), radius = 15, dtype = 'uint16'):
    """
    grabs a square box around center with given radius. Note that first entry in center is x coordinate (or cols) and second is y (rows)

    Example: center = (100,100) and radirus = 15.
    return array will contain data with shape (31,31) centered at pixel (100,100).
    """
    from numpy import nan,zeros,array
    x, y = center
    r = radius
    flag1 = ((x+r) < arr.shape[1])
    flag2 = ((y+r) < arr.shape[0])
    flag3 = ((x-r) > 0)
    flag4 = ((y-r) > 0)
    if flag1*flag2*flag3*flag4:
        result = arr[y-r:y+r+1,x-r:x+r+1]
    else:
        result = zeros((2*r+1,2*r+1))
        for idx in range(2*r+1):
            for idy in range(2*r+1):
                isx = x-r+idx
                isy = y-r+idy
                if (isx < arr.shape[1]) and (isy < arr.shape[0]) and (isx >= 0) and (isy >= 0):
                    result[idy,idx] = arr[isy,isx]
                else:
                    result[idy,idx] = 0

    return result

def get_random_array(size = (3000,4096),range = (0,4094), dtype = 'uint16'):
    """
    returns random array
    """
    from numpy.random import randint
    if dtype == 'uint16':
        arr = randint(range[0],range[1],size = size,dtype = dtype)
    else:
        arr = randint(range[0],range[1],size = size,dtype = dtype)
    return arr

def return_noise(size = (10,10), mean = 15, variance = 2, dtype = 'uin16'):
    """
    returns an numpy array with given size, mean value and variance. If dtype
    """
    from numpy.random import normal
    from numpy import sqrt
    noise = normal(loc=mean, scale=sqrt(variance), size=size).astype(dtype)
    return noise

def get_histogram(arr,length = 16,step = 1):
    """
    assumes unsigned int 16
    """
    from numpy import arange, histogram
    bins = arange(0,length,step) #calculating histogram
    y,x = histogram(arr,bins = bins)
    return x[:-1],y

def gaussian2D_from_shape(shape = (100,100), amplitude = 3000, position = (100,100), sigma = (5,5), dtype = 'uint16'):
    """
    return 2D gaussian function in a given 'position' on the provided image. The input image can be made of all zeros.
    """
    from numpy import sqrt,exp,copy,zeros
    r_mu = position[0]
    c_mu = position[1]
    r_sigma = sigma[0]
    c_sigma = sigma[1]
    gaussian = zeros(shape = shape)
    for r in range(gaussian.shape[0]):
        for c in range(gaussian.shape[1]):
            gaussian[r,c] = amplitude*exp(-((r-r_mu)**2/(2.0*r_sigma**2))-( (c-c_mu)**2 /( 2.0 * c_sigma**2) ) )

    return gaussian.astype(dtype)

def bin_array(num, m):
    from numpy import uint8, binary_repr,array
    """Convert a positive integer num into an m-bit bit vector"""
    return array(list(binary_repr(num).zfill(m))).astype(uint8)

def binarr_to_number(vector):
    """
    converts a vector of bits into an integer.
    """
    num = 0
    from numpy import flip
    vector = flip(vector)
    length = vector.shape[0]
    for i in range(length):
       num += (2**(i))*vector[i]
    return num

def nonzeromax(arr):
    """
    returns non-zero and nan maximum of a given array.
    """
    from numpy import nanmax, where
    idx = where(arr != 0)
    if idx[0].shape[0] > 0:
        return nanmax(arr[idx])
    else:
        return None

def nonzeromin(arr):
    """
    returns non-zero and nan minimum of a given array.
    """
    from numpy import nanmin, where
    idx = where(arr != 0)
    if idx[0].shape[0] > 0:
        return nanmin(arr[idx])
    else:
        return None

def gaussian1D(x, amp, x0, sigma, offset):
    """
    simple one-dimensional gaussian function
    """
    from numpy import exp
    y = amp*exp(-(x-x0)**2/(2*sigma*sigma))
    return y

def gaussian2D_from_mesh(mesh, amplitude, x0, y0, x_sigma, y_sigma, offset = 0 , theta = 0):
    """
    returns two-dimensional gaussian

    Parameters
    ----------
    mesh (2d numpy array)
    amplitude (float)
    x0 (float)
    y0 (float)
    x_sigma (float)
    y_sigma  (float)
    offset (float)
    theta (float)

    Returns
    -------
    z (2d numpy array)

    Examples
    --------
    >>> x = np.linspace(0, 20, 21)
    >>> y = np.linspace(0, 20, 21)
    >>> x,y = np.meshgrid(x, y)
    >>> xy = (x,y)
    >>> amp, x0, y0, sigmax, sigmay,offset, theta = 100,10,10,3,3,0,0
    >>> z = gaussian2D_from_mesh(xy,amp,x0,y0,sigmax,sigmay,offset)
    """
    from numpy import cos, sin, exp
    x = mesh[0]
    y = mesh[1]
    a = (cos(theta)**2)/(2*x_sigma**2) + (sin(theta)**2)/(2*y_sigma**2)
    b = -(sin(2*theta))/(4*x_sigma**2) + (sin(2*theta))/(4*y_sigma**2)
    c = (sin(theta)**2)/(2*x_sigma**2) + (cos(theta)**2)/(2*y_sigma**2)
    z = offset + amplitude*exp( - (a*((x-x0)**2) + 2*b*(x-x0)*(y-y0) + c*((y-y0)**2)))
    return z

def noise(arr,mean,sigma):
    """
    returns normal distributed noise array
    of shape x with mean and sigma.
    """
    from numpy.random import normal
    result = normal(mean,sigma,arr.shape)
    return result

def pixelate_xy(x,y,pixel_length = 10, dtype = None, saturation_value = None):
    """
    returns pixelated x,y-data. The length of x and y has to be divisable by pixel_size.
    """
    from numpy import zeros
    x_len = x.shape[0]
    y_len = y.shape[0]

    if x_len != y_len:
        raise ValueError('The shape of x and y input vectors has to be the same')

    if (x_len%pixel_length)!=0 or (y_len%pixel_length)!=0:
        raise ValueError('The length of the input data arrays has to be divisable by pixel_length')

    if dtype is None:
        x_dtype = x.dtype
        y_dtype = y.dtype
    else:
        x_dtype = dtype
        y_dtype = dtype

    x_new_len = int(x_len/pixel_length)
    y_new_len = int(y_len/pixel_length)
    x_new = zeros((x_new_len,), dtype = x_dtype)
    y_new = zeros((y_new_len,), dtype = y_dtype)
    for i in range(int(x_new_len)):
        x_new[i] = x[i*pixel_length:(i+1)*pixel_length].mean()
        y_new[i] = y[i*pixel_length:(i+1)*pixel_length].mean()
    if saturation_value is not None:
        y_new[y_new>=saturation_value] = saturation_value
    return x_new,y_new

def pixelate_image(x,y,z,pixel_size = 10, saturation_value = None):
    """
    returns pixilated image with pixel_size as input.
    The shape has to be divisible by pixel_size

    .. math::

        z =

    Parameters
    ----------
    mesh (2d numpy array)
    amplitude (float)
    x0 (float)
    y0 (float)
    x_sigma (float)
    y_sigma  (float)
    offset (float)
    theta (float)

    Returns
    -------
    z (2d numpy array)

    Examples
    --------
    >>> x = np.linspace(0, 20, 21)
    >>> y = np.linspace(0, 20, 21)
    >>> x,y = np.meshgrid(x, y)
    >>> xy = (x,y)
    >>> amp, x0, y0, sigmax, sigmay,offset, theta = 100,10,10,3,3,0,0
    >>> z = gaussian2D_from_mesh(xy,amp,x0,y0,sigmax,sigmay,offset)
    """
    from numpy import zeros
    x_shape = x.shape[0]
    y_shape = y.shape[0]
    z_shape = z.shape[0]
    x_new_len = int(x_shape/pixel_size)
    y_new_len = int(y_shape/pixel_size)
    z_new_len = int(z_shape/pixel_size)
    x_new = zeros((x_new_len,x_new_len))
    y_new = zeros((y_new_len,y_new_len))
    z_new = zeros((z_new_len,z_new_len))
    for i in range(int(x_new_len)):
        for j in range(int(x_new_len)):
            y_new[i,j] = y[i*pixel_size:(i+1)*pixel_size,j*pixel_size:(j+1)*pixel_size].mean()
            x_new[i,j] = x[i*pixel_size:(i+1)*pixel_size,j*pixel_size:(j+1)*pixel_size].mean()
            z_new[i,j] = z[i*pixel_size:(i+1)*pixel_size,j*pixel_size:(j+1)*pixel_size].mean()
    if saturation_value is not None:
        z_new[z_new>=saturation_value] = saturation_value
    return x_new,y_new,z_new

def nearest_neibhour(row,col):
    """
    returns an matrix of indices where fast axis (axis = 0) corresponds to particle index and the resulting vector show the order of nearest neibhours.


    Parameters
    ----------
    row (1d numpy array)
    col (1d numpy array)
    Returns
    -------
    matrix (2d numpt array)

    Examples
    --------
    >>> import numpy as np
    >>> row = np.asarray([100, 200,400, 600, 150])
    >>> col = np.asarray([100, 200,300, 400, 300])
    >>> nn = nearest_neibhour(row = row,col = col)
    >>> nn
    array([[0, 1, 4, 2, 3],
       [1, 4, 0, 2, 3],
       [2, 1, 3, 4, 0],
       [3, 2, 1, 4, 0],
       [4, 1, 0, 2, 3]])

    >>> print("and visualy can be represented as a scatter chart, where the marker is replaced with a closest neibhours. 0 - stands for the point for which neibhours are calculated.")"

    .. plot::

        >>> import matplotlib.pyplot as plt
        >>> import numpy as np
        >>> from ubcs_auxiliary import numerical
        >>> row = np.asarray([100, 200,400, 600, 150])
        >>> col = np.asarray([100, 200,300, 400, 300])
        >>> nn = numerical.nearest_neibhour(row = row, col = col)
        >>> plt.grid()
        >>> for i in range(len(nn[1])):
        >>>    plt.scatter(col[nn[1,i]],row[nn[1,i]], s=400, marker = f'${i}$')
        >>> plt.title('example: point 1')
        >>> plt.show()

    """
    from numpy import zeros, nan, nanmin, amin, ones, sqrt, where, argsort

    matrix = distance_matrix(row,col)
    matrix_argsort = argsort(matrix,axis=1)
    return matrix_argsort

def distance_matrix(row,col):
    """
    returns a matrix of all pair-wise distances.

    Parameters
    ----------
    row (1d numpy array)
    col (1d numpy array)
    Returns
    -------
    matrix (2d numpt array)

    Examples
    --------
    >>> import numpy as np
    >>> row = np.asarray([100,200,400])
    >>> col = np.asarray([100,200,300])
    >>> dist = distance_matrix(row = row,col = col)
    >>> dist
    array([[  0.        , 141.42135624, 360.55512755],
       [141.42135624,   0.        , 223.60679775],
       [360.55512755, 223.60679775,   0.        ]])
    """
    from numpy import zeros, meshgrid,sqrt
    # vectorized form, might use more RAM
    row_i, row_j = meshgrid(row, row, sparse=True)
    row_m = ((row_i-row_j)**2)
    col_i, col_j = meshgrid(col, col, sparse=True)
    col_m = ((col_i-col_j)**2)
    matrix = sqrt(row_m + col_m)
    del col_m, row_m, row_i, row_j,col_i, col_j
    return matrix

def linear_coeff_from_points(p1,p2):
    """
    p1(x1,y1)
    p2(x2,y2)

    y = a*x+b
    """
    y1 = p1[1]
    y2 = p2[1]
    x1 = p1[0]
    x2 = p2[0]

    a = (y1-y2)/(x1-x2)
    b = (x1/(x1-x2))*(y1-y2) + y1
    return a,b

def max_filter(image, footprintsize = 10, treshhold = 100):
    """
    returns mask with max values with given distance

    to find the mask of max pixels. These are peak pixels in the image.
    """
    from scipy.ndimage import maximum_filter
    from numpy import ones
    footprint = ones((footprintsize,footprintsize))
    max_mask = (maximum_filter(image,footprint = footprint) == image)&(image>=treshhold)

    return max_mask



#!/usr/bin/python
# -*- coding: utf-8 -*-
"""
plotting functions
author: Valentyn Stadnytskyi
data: 2017 - Nov 17 2018

"""
__vesrion__ = '1.0.0'
from time import  time, sleep
import sys
if sys.version_info[0] == 2:
    from time import clock as perf_counter
else:
    from time import perf_counter
import platform
from ubcs_auxiliary.save_load_object import save_to_file, load_from_file

#def generate_plot(x,y):#
    #from matplotlib.figure import Figure
    #from matplotlib import pyplot
    #from matplotlib.backends.backend_wxagg import FigureCanvasWxAgg
    #figure = Figure(figsize=(8,3),dpi=80)
    #axis = figure.add_subplot(1,1,1)
    #axis.plot(x,y)
    #return axis

def generate_plot(x,y):
    from matplotlib.figure import Figure
    from matplotlib import pyplot as plt
    #from matplotlib.backends.backend_wxagg import FigureCanvasWxAgg
    figure = plt.figure()
    axis = figure.add_subplot(1,1,1)
    axis.plot(x,y)
    return figure

def plot_frames_as_RGB(frames = (None,None,None)):
    """
    takes a tuple of 3 mono-color frames and plots them as RGB. where RGB encoded the order of the supplied frames.

    All frames have to be the same size. Currently works for 3000x4096 size only.


    """
    from matplotlib import pyplot as plt
    from numpy import zeros, ones
    img = ones(frames[0].shape+(3,))
    for i in range(3):
        img[:,:,i] = frames[i]
    fig = plt.figure(figsize=(4, 4))
    grid = plt.GridSpec(1, 1, hspace=0.025, wspace=0.025)
    ax1 = fig.add_subplot(grid[0,0])
    ax1.imshow(img)

def plot_image_with_mask(image,mask):
    """
    plot an image and overlay a mask over it. The mask shows
    """
    from matplotlib import pyplot as plt
    from numpy import where
    fig = plt.figure(figsize=(4, 4))
    grid = plt.GridSpec(1, 1, hspace=0.025, wspace=0.025)
    ax1 = fig.add_subplot(grid[0,0])
    ax1.imshow(image)
    #fig.colorbar(cax = ax1,orientation='vertical', pad=0.0, shrink=0.855)
    xy_coordinates = where(mask)
    ax1.plot(xy_coordinates[1],xy_coordinates[0],'wo',markersize=3)

def plot_images(lst = [], vmax = None, vmin = None, titles = None):
    from matplotlib import pyplot as plt

    if vmax is None:
        vmax = []
        for item in lst:
            vmax.append(None)
    if vmin is None:
        vmin = []
        for item in lst:
            vmin.append(None)
    if titles is None:
        titles = []
        for item in lst:
            titles.append(None)

    fig = plt.figure(figsize=(7, 3))
    grid = plt.GridSpec(1, len(lst), hspace=0.025, wspace=0.025)
    ax = []
    ax.append(fig.add_subplot(grid[0,0]))
    ax[0].imshow(lst[0],vmax = vmax[0], vmin = vmin[0])
    ax[0].set_title(titles[0])
    ax[0].set_axis_off()
    for i in range(1,len(lst)):
        ax.append(fig.add_subplot(grid[0,i], sharex = ax[0], sharey = ax[0]))
        ax[i].imshow(lst[i],vmax = vmax[i], vmin = vmin[i])
        ax[i].set_title(titles[i])
        ax[i].set_axis_off()

def plot_images_grid(lst = [], vmax = None, vmin = None, titles = None):
    from matplotlib import pyplot as plt

    if vmax is None:
        vmax = []
        for item in lst:
            vmax.append(None)
    if vmin is None:
        vmin = []
        for item in lst:
            vmin.append(None)
    if titles is None:
        titles = []
        for item in lst:
            titles.append(None)

    fig = plt.figure(figsize=(7, 3))
    grid = plt.GridSpec(1, len(lst), hspace=0.025, wspace=0.025)
    ax = []
    ax.append(fig.add_subplot(grid[0,0]))
    ax[0].imshow(lst[0],vmax = vmax[0], vmin = vmin[0])
    ax[0].set_title(titles[0])
    ax[0].set_axis_off()
    for i in range(1,len(lst)):
        ax.append(fig.add_subplot(grid[0,i], sharex = ax[0], sharey = ax[0]))
        ax[i].imshow(lst[i],vmax = vmax[i], vmin = vmin[i])
        ax[i].set_title(titles[i])
        ax[i].set_axis_off()

def grid_plot_3d_example():
    import matplotlib.pyplot as plt
    from matplotlib import cm
    import numpy as np

    from mpl_toolkits.mplot3d.axes3d import get_test_data
    # This import registers the 3D projection, but is otherwise unused.
    from mpl_toolkits.mplot3d import Axes3D  # noqa: F401 unused import


    # set up a figure twice as wide as it is tall
    fig = plt.figure(figsize=plt.figaspect(0.5))
    grid = plt.GridSpec(4, 4, hspace=0.025, wspace=0.025)
    #===============
    #  First subplot
    #===============
    # set up the axes for the first plot
    ax1 = fig.add_subplot(grid[0:3,0:3], projection='3d')

    # plot a 3D surface like in the example mplot3d/surface3d_demo
    X = np.arange(-5, 5, 0.25)
    Y = np.arange(-5, 5, 0.25)
    X, Y = np.meshgrid(X, Y)
    R = np.sqrt(X**2 + Y**2)
    Z = np.sin(R)
    surf = ax1.plot_surface(X, Y, Z, rstride=1, cstride=1, cmap=cm.coolwarm,
                           linewidth=0, antialiased=False)
    ax.set_zlim(-1.01, 1.01)
    fig.colorbar(surf, shrink=0.5, aspect=10)

    #===============
    # Second subplot
    #===============
    # set up the axes for the second plot
    ax2 = fig.add_subplot(grid[3,3], projection='3d')

    # plot a 3D wireframe like in the example mplot3d/wire3d_demo
    X, Y, Z = get_test_data(0.05)
    ax2.plot_wireframe(X, Y, Z, rstride=10, cstride=10)



    plt.show()

def example2(fig):
    from matplotlib import pyplot as plt
    fig = plt.figure()
    ax1 = fig.add_subplot(121)
    ax2 = fig.add_subplot(122)

if __name__ == '__main__':
    from numpy import arange, random
    #x = arange(0,100,1)
    #y = random.random((100,))
    #save_object('plot.pltpkl',generate_plot(x,y))
    #figure_loaded = load_object('plot.pltpkl')

    #figure_loaded.show()



def find(topdir, name=[], exclude=[]):
    """A list of files found starting at 'topdir' that match the patterns given
    by 'name', excluding those matching the patterns given by 'exclude'.

    Parameters
    ----------
    topdir (string)
    name (list)
    exclude (list)

    Returns
    -------
    file_list (list)

    Examples
    --------
    >>> res = anfinrud_auxiliary.os.walk('ubcs_auxiliary/')
    >>> for i in res: print(i[0])
            ...:
        ubcs_auxiliary/
        ubcs_auxiliary/tests
        ubcs_auxiliary/tests/__pycache__
        ubcs_auxiliary/__pycache__

    Python 3.7.4 (v3.7.4:e09359112e, Jul  8 2019, 14:54:52)
    Type 'copyright', 'credits' or 'license' for more information
    IPython 7.8.0 -- An enhanced Interactive Python. Type '?' for help.
    In [1]: from time import time
    In [2]: import auxiliary
    In [3]: res = auxiliary.os.walk('/')
    In [4]: t1 = time(); lst = list(res); t2 = time(); print(t2-t1, len(lst))
    3.9815242290496826 1346

    Python 2.7.16 (v2.7.16:413a49145e, Mar  2 2019, 14:32:10)
    Type "copyright", "credits" or "license" for more information.
    IPython 5.8.0 -- An enhanced Interactive Python.
    In [1]: from time import time
    In [2]: import anfinrud_auxiliary
    In [3]: res = auxiliary.os.walk('')
    In [4]: t1 = time(); lst = list(res); t2 = time(); print(t2-t1, len(lst))
    (0.77646803855896, 1346)
    """

    def glob_to_regex(pattern):
        return "^"+pattern.replace(".", "\\.").replace("*", ".*").replace("?", ".")+"$"
    try:
        from scandir import walk
    except ImportError:
        from os import walk
    import re
    if type(name) == str:
        name = [name]
    if type(exclude) == str:
        exclude = [exclude]
    name = [re.compile(glob_to_regex(pattern)) for pattern in name]
    exclude = [re.compile(glob_to_regex(pattern)) for pattern in exclude]

    file_list = []
    for (directory, subdirs, files) in walk(topdir):
        for file in files:
            pathname = directory+"/"+file
            match = any([pattern.match(pathname) for pattern in name]) and\
                not any([pattern.match(pathname) for pattern in exclude])
            if match:
                file_list += [pathname]
    return file_list

def exclude():
    """Returns a list of patterns to exclude from a search. Add
    terms as required."""
    exclude = ['*/alignment*',
               '*/trash*',
               '*/_Archived*',
               '*/backup*',
               '*/Commissioning*',
               '*/Test*',
               '*/.AppleDouble*',
               '*LaserX*',
               '*LaserZ*',
               '*Copy*',
               '*._*',
               '.DS_Store']
    return exclude

def image_file_names_from_path(beamtime,path_name):
    """Returns image file names found under path_name. Typically used with
    'Reference_*() , which specifies which directories contain data for which
    zinger-free-statistics are to be acquired. The zinger-free-statistics
    include Imean and Ivar, which are used to construct UC_psi.npy."""
    from db_functions import find,exclude
    from numpy import sort
    data_dir,analysis_dir = data_dir_analysis_dir(beamtime)
    terms = ['*.mccd','*.rx']
    image_files = sort(find(data_dir+path_name, name=terms, exclude=exclude()))
    return image_files

def N_files_in_dir(folder = '', match = '*'):
    import os
    import fnmatch
    integer = len(fnmatch.filter(os.listdir(folder), match))
    return integer

def listdir(root, include = ['.hdf5'], exclude = [], sort = ''):
    """
    returns list of files from a 'source' directory that have terms listed in 'include' and doesn't terms listed in 'exclude'. Extra parameter 'sort' can be used to sort the output list. If left blank, no sorting will be performed.

    Parameters
    ----------
    source (string)
    include (list)
    exclude (list)
    sort (string)

    Returns
    -------
    file_list (list)

    Examples
    --------
    >>> res = ubcs_auxiliary.os.get_list_of_files('/',['.hdf5'])
    """
    import os
    from numpy import argsort, array
    files = [os.path.join(root,file) for file in os.listdir(root)]
    selected = []
    [selected.append(file) for file in files if ((all([term in file for term in include])) and (all([term2 not in file for term2 in exclude])))]

    return selected

def find_recent_filename(root, include, exclude, newest_first = True):
    """
    find the list of files or folders that have any terms specified in the list 'include' and do not have terms specified in 'exclude'. The extra parameter reverse_order specified whether return the newest or oldest one.

    Parameters
    ----------
    source (string)
    include (list)
    exclude (list)
    sort (string)

    Returns
    -------
    file_list (list)
    """
    from os import listdir,path
    from os.path import getmtime
    from numpy import argsort, array
    files = [path.join(root,file) for file in listdir(root)]
    selected = []
    [selected.append(file) for file in files if ((all([term in file for term in include])) and (all([term2 not in file for term2 in exclude])))]
    path_names = selected.copy()
    if len(path_names) > 0:
        creation_times = [getmtime(file) for file in path_names]
        sort_order = argsort(creation_times)
        if newest_first:
            return array(path_names)[sort_order][-1]
        else:
            return array(path_names)[sort_order][0]
    else:
        return ''

def get_current_pid_memory_usage(units = 'GB',verbose = False):
    """
    returns current process memory footprint.
    """
    import os
    import psutil
    pid = os.getpid()
    py = psutil.Process(pid)
    coeff = 1
    if units == 'GB':
        coeff = 2.**30
    elif units == 'MB':
        coeff = 2.**20
    elif units == 'KB':
        coeff = 2.**10
    elif units == 'B':
        coeff = 2.**0.0
    elif units == 'b':
        coeff = 1.0/8.0
    memoryUse = py.memory_info()[0]/coeff  # memory use in GB...I think
    if verbose:
        print('memory use:', memoryUse)
    else:
        pass
    return memoryUse

def does_filename_have_counterpart(src_path,dst_root = None, counterpart_extension = ''):
    """
    checks if the 'src_path' has counterpart with extension 'counterpart_extension'.

    Parameters
    ----------
    filename (string)

    counterpart_extension (string)

    Returns
    -------
    boolean (boolean)
    """
    import os
    src_root, src_name = os.path.split(src_path)
    if dst_root is None:
        dst_root, dst_name = os.path.split(src_path)

    splitted = src_name.split('.')
    src_base = splitted[0]
    src_extension = ''.join(splitted[1:])
    counterpart = os.path.join(dst_root,src_base + counterpart_extension)
    flag = os.path.exists(counterpart)
    return flag


def read_config_file(filename):
    """
    read yaml config file

    Parameters
    ----------
    filename (string)

    Returns
    -------
    dict (dictionary)
    boolean (boolean)
    """
    import yaml
    import os
    flag =  os.path.isfile(filename)
    if flag:
        with open(filename,'r') as handle:
            config = yaml.safe_load(handle.read())  # (2)
    else:
        config = {}
    return config, flag



from numpy import zeros, array_equal

from ubcs_auxiliary.save_load_object import save_to_file,  load_from_file

def test_save_load(tmpdir):
    import os
    print(f'tempdir = {tmpdir}')
    print(f'{os.path.exists(tmpdir)}')
    arr_in = zeros((10,5))
    str_in = 'this is string'
    dict_in = {'1':1,'2':2}
    list_in = [1,2,3,4]
    compound_in = {}
    compound_in['arr'] = arr_in
    compound_in['str'] = str_in
    compound_in['dict'] = dict_in
    compound_in['list'] = list_in

    # save to a file
    save_to_file(tmpdir+'/arr.pkl',arr_in)
    save_to_file(tmpdir+'/str.pkl',str_in)
    save_to_file(tmpdir+'/dict.pkl',dict_in)
    save_to_file(tmpdir+'/list.pkl',list_in)
    save_to_file(tmpdir+'/compound.pkl',compound_in)

    # Load to a file
    arr_out = load_from_file(tmpdir+'/arr.pkl')
    str_out = load_from_file(tmpdir+'/str.pkl')
    dict_out = load_from_file(tmpdir+'/dict.pkl')
    list_out = load_from_file(tmpdir+'/list.pkl')
    compound_out = load_from_file(tmpdir+'/compound.pkl')

    #run tests
    assert array_equal(arr_in, arr_out)
    assert str_in == str_out
    assert dict_in == dict_out
    assert list_in == list_out
    assert compound_in.keys() == compound_out.keys()

def test_save_load_hdf5(tmpdir):
    from ubcs_auxiliary.save_load_object import save_to_hdf5,  load_from_hdf5
    import os
    print(f'tempdir = {tmpdir}')
    print(f'{os.path.exists(tmpdir)}')
    arr_in = {'array':zeros((10,5))}
    str_in = {'string':b'this is string'}
    dict_in = {'1':1,'2':2}
    list_in = {'list':[1,2,3,4]}
    compound_in = {}
    compound_in['array'] = arr_in['array']
    compound_in['string'] = str_in['string']
    compound_in['list'] = list_in['list']

    # save to a file
    save_to_hdf5(tmpdir+'/arr.hdf5',arr_in)
    save_to_hdf5(tmpdir+'/str.hdf5',str_in)
    save_to_hdf5(tmpdir+'/dict.hdf5',dict_in)
    save_to_hdf5(tmpdir+'/list.hdf5',list_in)
    save_to_hdf5(tmpdir+'/compound.hdf5',compound_in)

    # Load to a file
    arr_out = load_from_hdf5(tmpdir+'/arr.hdf5')
    str_out = load_from_hdf5(tmpdir+'/str.hdf5')
    dict_out = load_from_hdf5(tmpdir+'/dict.hdf5')
    list_out = load_from_hdf5(tmpdir+'/list.hdf5')
    compound_out = load_from_hdf5(tmpdir+'/compound.hdf5')

    #run tests
    assert array_equal(arr_in['array'], arr_out['array'])
    assert str_in == str_out
    assert dict_in == dict_out
    print('lists',list_in,list_out)
    #Note HDF5 returns a saved list as an array
    assert list_in['list'] == list(list_out['list'])
    assert compound_in.keys() == compound_out.keys()



def compress_file(filename, compressoin = 0, retain = False):
    """
    a function that recompresses the hdf5 file.
    """
    from time import ctime, time
    import os


def compress_files(dir, compressoin = 0, retain = False):
    import os
    from time import time, ctime
    src_lst = os.listdir()
    dest_lst = []
    for item in src_lst:
        if retain:
            flag = '.hdf5' in item  and '.hdf5.gzip9' not in item
        else:
            flag = '.hdf5' in item
        if flag:
            dest_lst.append(item)
    print(dest_lst)
    for item in dest_lst:
        src_file = item
        dst_file = item+'.gzip9'
        print('start',ctime(time()))
        os.system(f'h5repack -v {src_file} {dst_file} GZIP=9')
        if not retain:
            os.remove(src_file)
        print('end',ctime(time()))



from ubcs_auxiliary.multithreading import new_thread

def test_start_new_thread():
    print('1')
    def func(*args,**kwargs):
        global sum_value
        global kwargs_list
        from numpy import sum
        sum_value = sum(args)
        kwargs_list = list(kwargs.keys())

    from time import sleep
    global sum_value
    global kwargs_list
    sum_value = None
    kwargs_list = None
    thread = new_thread(func,1,2,3,4,5,6 , keyword1 = '', keyword2 = 5)
    while kwargs_list is None:
        sleep(0.01)
    assert sum_value == 21
    assert kwargs_list == ['keyword1','keyword2']
print('2')
def test_start_new_thread2():
    from time import sleep
    global sum_value
    global kwargs_list
    def func(*args,**kwargs):
        global sum_value
        global kwargs_list
        from numpy import sum
        sum_value = sum(args)
        kwargs_list = list(kwargs.keys())
    sum_value = None
    kwargs_list = None
    new_thread(func,-1,-2,-3,-4,-5,-6, keyword12 = '', keyword22 = 5)
    while kwargs_list is None:
        sleep(0.01)
    assert sum_value == -21
    assert kwargs_list == ['keyword12','keyword22']
print('3')
def test_start_new_thread3():
    from time import sleep
    global value
    value = 0
    def func():
        global value
        value = 5
    new_thread(func)

    sleep(0.01)
    assert value == 5



"""
Traveling Salesman problem for 2d (x,y) case
author: Valentyn Stadnytskyi
data: 2017 - Nov 17 2018

The traveling salesman algorithm takes an input array of coordinates and vary array (True/False)

"""
__vesrion__ = '1.0.0'
from time import  time, sleep
import sys
if sys.version_info[0] == 2:
    from time import clock as perf_counter
else:
    from time import perf_counter
import platform


def create_test_sequence(N, dim = 2):
    """

    """
    from numpy import random, array
    arr = (random.rand(N, dim)-0.5)*10
    arr[0] = array([0]*dim)
    arr[-1] = array([0]*dim)
    vary = arr[:,0]*0+1
    vary[0] = 0
    vary[-1] = 0
    return arr, vary

def euclidean_distance(i,j):
    """
    return euclidean_distance between two points
    """
    from numpy import array, nan
    dim_i = len(i)
    dim_j = len(j)
    if dim_i == dim_j:
        dim = dim_i
        distance = ((array(i) - array(j))**2).sum()**0.5
    else:
        distance = nan
    return distance

def total_distance(arr):
    """
    returns total distance as a sum of distances.
    """
    length = arr.shape[0]
    sum_distance = 0.0
    for i in range(length-1):
        a = arr[i]
        b = arr[i+1]
        sum_distance += euclidean_distance(a,b)
    return sum_distance

def distance(i,j):
    """
    simple wrapper around euclidean_distance
    """
    return euclidean_distance(i,j)

def minimize(arr, vary,  method = 'brute_force', N = 300):
    if method == 'brute_force':
        for i in range(int(N)):
            arr = minimize_brute_force_once(arr)
        return arr
    elif method == 'temperature':
        return arr

def minimize_brute_force_once(arr):
    """
    complete random minimization algorithm; slow and inefficient
    """
    from numpy.random import shuffle
    from numpy import concatenate, copy
    mid_array = arr[1:-1]
    new_mid_array = copy(mid_array)

    new_arr = copy(arr)
    new_arr[0] = arr[0]
    new_arr[-1] = arr[-1]

    shuffle(new_mid_array)
    new_arr[1:-1] = new_mid_array

    distance = total_distance(arr)
    new_distance = total_distance(new_arr)
    if new_distance < distance:
        return new_arr
    else:
        return arr

def minimize_temperature(self):
    """
    """
    import random, numpy, math, copy
    tour = list(self.sequence)
    cities = self.data
    N = len(self.sequence)
    for temperature in numpy.logspace(0,5,num=100000)[::-1]:
        [i,j] = sorted(random.sample(list(range(N)), 2));
        newTour =  tour[:i] + tour[j:j + 1] +  tour[i + 1:j] + tour[i:i + 1] + tour[j+1:];
        if math.exp((sum([ math.sqrt(sum([(cities[tour[(k+1) % N]][d] - cities[tour[k % N]][d])**2 for d in [0,1] ])) for k in [j,j-1,i,i-1]]) - sum([math.sqrt(sum([(cities[newTour[(k+1) % N]][d] - cities[newTour[k % N]][d])**2 for d in [0,1] ])) for k in [j,j-1,i,i-1]])) / temperature) > random.random():
            self.sequence = copy.copy(newTour);

def init(data):
    from numpy import array
    self.data = data
    self.length = int(data.size/2)
    self.sequence = range(1, self.length)

def run(data):
    self.init(data)
    self.minimize(N=500)
    return self.sorted_data

@property
def sorted_data(self):
    from numpy import array
    data = []
    data.append(self.data[0])
    for item in self.sequence:
        data.append(self.data[item])
    data.append(self.data[0])
    return array(data)






##### OLD functions
def matrix(arr):
    from numpy import zeros, roll, nan, nanmax, arange
    length = int(arr.size/2)
    N_vector = arange(0, length,1)
    x_vector = arr[:, 0]
    y_vector = arr[:, 1]
    Mx = zeros((length, length))
    My = zeros((length, length))
    Mn = zeros((length, length))
    for i in range(length):
        Mx[i] = roll(x_vector, i)
        My[i] = roll(y_vector, i)
        Mn[i] = roll(N_vector, i)
    return Mx, My, Mn

def minimize_matrix(Mx, My, Mn):
    from numpy import matmul, nan, where, nanmax
    Dx = matmul(Mx, Mx.T)
    Dy = matmul(My, My.T)
    length = Mx.shape[0]
    D = Dx+Dy
    for i in range(length):
        D[i,i] = nan
    res = where(D == nanmax(D))
    return res

def test(self):
    global Mn
    N = 5
    self.init(N)
    print(self.minimize())
    Mx, My, Mn = self.matrix()
    print(Mn)
    self.plot_x()
    return Mn,Mx,My

def plot(arr, vary = None, labels = None):
    """
    visualizes the array of coordintes. Parameters vary and labels can be used to customize the graph. Do not work for now. TODO.
    """
    from matplotlib import pyplot as plt
    from numpy import where
    plt.ion()
    if vary is None:
        vary = arr[:,0]*0+1
        vary[0] = vary[-1] = 0
    x = arr[:,0]
    y = arr[:,1]
    x_data = x[where(vary != 0)]
    y_data = y[where(vary != 0)]
    x_center = x[where(vary == 0)]
    y_center = y[where(vary == 0)]
    if labels is None:
        n = list(range(len(x-2)))
    fig, ax = plt.subplots()
    ax.plot(x, y, '-o')
    for i, txt in enumerate(n):
        ax.annotate(txt, (x[i], y[i]))
    plt.show()

if __name__ == '__main__':
    from pdb import pm
    arr, vary = create_test_sequence(10)
    total_dist = total_distance(arr)



"""
advance Sleep functions
author: Valentyn Stadnytskyi
data: 2017 - Nov 17 2018

functions:
psleep - precision sleep takes
intsleep with inputs t, dt and interupt as a function.

The precision sleep class.
functiob:
psleep - sleep specified amount of time with sub milisecond precision
test_sleep - for testing purposes. will print how much time the code waited.
This is important for the Windows platform programs if precise wait is required.
The Windows OS has ~15-17 ms latenct - the shortest time between attentions from OS.
"""

def beep():
    from IPython.display import Audio
    sound_file = 'http://www.soundjay.com/button/beep-07.wav'
    Audio(sound_file, autoplay=True)



"""
advance Sleep functions
author: Valentyn Stadnytskyi
data: 2017 - Nov 17 2018

functions:
psleep - precision sleep takes
intsleep with inputs t, dt and interupt as a function.

The precision sleep class.
functiob:
psleep - sleep specified amount of time with sub milisecond precision
test_sleep - for testing purposes. will print how much time the code waited.
This is important for the Windows platform programs if precise wait is required.
The Windows OS has ~15-17 ms latenct - the shortest time between attentions from OS.
"""
__vesrion__ = '1.0.0'
from time import  time, sleep
import sys
if sys.version_info[0] == 2:
    from time import clock as perf_counter
else:
    from time import perf_counter
import platform



def precision_sleep(t = 0.02, min_time = 0.017):
    """
    sleep for t seconds.
    """
    from time import time, sleep,perf_counter
    import platform
    min_time

    time_start = perf_counter()
    if platform.system() == 'Linux':
        sleep(t)
    #elif platform.system() == 'Darwin':
    #    sleep(t)
    else:
        if t>min_time:
            sleep(t-min_time)
            time_left = t - (perf_counter() - time_start)
            time_while_start = perf_counter()
            while perf_counter() - time_while_start <= time_left:
                pass
        else:
            time_left = t - (perf_counter() - time_start)
            time_while_start = perf_counter()
            while perf_counter() - time_while_start <= time_left:
                pass

def interupt_sleep(t= 0.02,dt = 0.01, interupt = None):
    """precision sleep function with interupt capabilities
    input:t - time to sleep, dt - check intervals, interupt - interupt function that return boolean
    """
    from time import perf_counter,sleep,time
    import types
    if not isinstance(interupt,types.FunctionType):
        precision_sleep(t)
    else:
        time_start = perf_counter()
        while perf_counter() - time_start <= t:
            if interupt(): break
            if (perf_counter() - time_start) >= dt:
                precision_sleep(dt)
            else:
                precision_sleep(perf_counter() - time_start)



if __name__ == '__main__':


    def test_psleep(t = 0.01):
        """
        test_sleep t = 0.01 in seconds
        """
        from time import time, sleep,perf_counter
        t1 =  perf_counter()
        psleep(t)
        t2 = perf_counter()
        dt = t2-t1
        print(dt)

    def test_intsleep(t = 0.1, dt = 0.1, interupt = False):
        """
        test_sleep t = 0.01 in seconds
        """
        from time import time, sleep,perf_counter
        print(t,dt,interupt)
        t1 =  perf_counter()
        intsleep(t = t,dt = dt, interupt = interupt)
        t2 = perf_counter()
        dt = t2-t1
        print(dt)
    def interupt():
        """
        return value of a global variable flag
        """
        global flag
        return flag
    flag = True
    print('test_psleep(0.010) # in seconds')
    print('test_intsleep(0.10,0.01,interupt) # in seconds')
    print('intsleep(0.10,0.1) # in seconds')



#!/usr/bin/python
# -*- coding: utf-8 -*-
"""
author: Valentyn Stadnytskyi
created: Sept 9 2019

"""

def save_to_file(filename, object, protocol = 'pickle'):
    """
    save a python object to a file

    Parameters
    ----------
    Args:
        filename (string)
            the full path and filename
        object (python object)
            a python object

    Returns
    -------

    Examples
    --------
    the example of usage

    >>> save_to_file('test.pkl',[1,2,3])
    """
    if protocol =='pickle':
        from pickle import dumps
        data = dumps(object)
        with open(filename,"wb") as f:
            f.write(data)
    elif protocol == 'msgpack':
        import msgpack
        import msgpack_numpy as m
        data = msgpack.packb(object, default=m.encode)
        filename += '.msgpck'
        with open(filename,"wb") as f:
            f.write(data)
    elif protocol == 'hdf5':
        from h5py import File
        filename += '.hdf5'
        with File(filename, "w") as file:
            file.create_dataset("object", data=object)


def load_from_file(filename, protocol = 'pickle'):
    """
    read object from a file

    Parameters
    ----------
    filename : string
        the full path and filename

    Returns
    -------
    object : object
        input object to save.

    Examples
    --------
    the example of usage

    >>> list_out = load_from_file('list.extension')

    """
    if protocol == 'pickle':
        from pickle import load
        with open(filename,'rb') as f:
            data = load(f, encoding='bytes')
        return data
    elif protocol == 'hdf5':
        from h5py import File
        with File(filename, "r") as f:
            pass


def save_to_hdf5(filename, dic_in, compression = 0):
    """
    save a dictionary to a .hdf5 file

    Parameters
    ----------
    filename (string)
        the full path and filename
    dict (dictionary)
        a python dictionary
    compression (integer)
        compression paramter

    Returns
    -------

    Examples
    --------
    the example of usage

    >>> save_to_hdf5('list.hd5f',[1,2,3])
    """
    from h5py import File
    from numpy import ndarray
    with File(filename, "w") as file:
        for key in list(dic_in.keys()):
            object = dic_in[key]
            if type(object) == ndarray:
                if compression > 0:
                    file.create_dataset(key, data=object, compression = 'gzip', compression_opt = compression)
                else:
                    file.create_dataset(key, data=object)
            else:
                file.create_dataset(key, data=object)


def load_from_hdf5(filename):
    """
    read object from a file

    Parameters
    ----------
    filename : string
        the full path and filename

    Returns
    -------
    object : object
        input object to save.

    Examples
    --------
    the example of usage

    >>> data_hdf5 = load_from_hdf5('list.extension.hdf5')

    """
    from h5py import File
    from numpy import copy
    data = {}
    with File(filename, "r") as f:
        for key in list(f.keys()):
            data[key] = f[key][()]
    return data
